#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import pint
import warnings

import numpy as np

from .. import config, core

from typing import Sequence

# Set up logging
logger = config.get_logger(__name__)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    pint.Quantity([])


def find_soma(x: 'core.TreeNeuron') -> Sequence[int]:
    """Try finding a neuron's soma.

    Will use the `.soma_detection_radius` and `.soma_detection_label`
    attribute of a neuron to search for the soma in the node table.

    If attributes don't exists, will fallback to defaults: `None` and
    `1`, respectively.

    Parameters
    ----------
    x :         Neuron

    Returns
    -------
    Node ID(s) of potential somata.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> navis.find_soma(n)
    array([4177], dtype=int32)

    """
    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Input must be TreeNeuron, not "{type(x)}"')

    soma_radius = getattr(x, 'soma_detection_radius', None)
    soma_label = getattr(x, 'soma_detection_label', None)

    check_labels = not isinstance(soma_label, type(None)) and 'label' in x.nodes.columns
    check_radius = not isinstance(soma_radius, type(None))

    # If no label or radius is given, return empty array
    if not check_labels and not check_radius:
        return np.array([], dtype=x.nodes.node_id.values.dtype)

    # Note to self: I've optimised the s**t out of this function
    # The reason reason why we're using a mask and this somewhat
    # convoluted logic is to avoid having to subset the node table
    # because that's really slow.

    # Start with a mask that includes all nodes
    mask = np.ones(len(x.nodes), dtype=bool)

    if check_radius:
        # When checking for radii, we use an empty mask and fill it
        # with nodes that have a large enough radius
        mask[:] = False

        # Drop nodes that don't have a radius
        radii = x.nodes.radius.values
        has_radius = ~np.isnan(radii)

        # Filter further to nodes that have a large enough radius
        if has_radius.any():
            if isinstance(soma_radius, pint.Quantity):
                if isinstance(x.units, (pint.Quantity, pint.Unit)) and \
                   not x.units.dimensionless and \
                   not isinstance(x.units._magnitude, np.ndarray) \
                   and x.units != soma_radius:  # only convert if units are different
                    # Do NOT remove the .values here -> otherwise conversion to units won't work
                    is_large = radii * x.units >= soma_radius
                else:
                    # If neurons has no units or if units are non-isotropic,
                    # assume they are the same as the soma radius
                    is_large = radii >= soma_radius._magnitude
            else:
                is_large = radii >= soma_radius

            # Mark nodes that have a large enough radius
            mask[is_large] = True

    # See if we (also) need to check for a specific label
    if check_labels:
        # Important: we need to use np.asarray here because the `label` column
        # can be categorical in which case a `soma_nodes.label.astype(str)` might
        # throw annoying runtime warnings
        soma_node_ids = x.nodes.node_id.values[mask]
        soma_node_labels = np.asarray(x.nodes.label.values[mask]).astype(str)

        return soma_node_ids[soma_node_labels == str(soma_label)]
    # If no labels to check we can return the mask directly
    else:
        return x.nodes.node_id.values[mask]




#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
"""Module containing functions and classes to build `NEURON` network models."""

import matplotlib.pyplot as plt
import matplotlib.colors as mcl
import numpy as np
import pandas as pd
import seaborn as sns

import plotly.graph_objects as go

from collections import namedtuple
from matplotlib.collections import LineCollection
from numpy.lib.stride_tricks import sliding_window_view

from ... import config, utils

# We will belay any import error
try:
    import neuron
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        'This interface requires the `neuron` libary to be '
        'installed:\n pip3 install neuron\n'
        'See also https://neuron.yale.edu/neuron/')

from neuron.units import ms, mV
neuron.h.load_file('stdrun.hoc')

# Set up logging
logger = config.get_logger(__name__)

__all__ = []

# It looks like there can only ever be one reference to the time
# If we have multiple models, we will each reference them to this variable
main_t = None

Stimulus = namedtuple('Stimulus', ['start', 'stop', 'frequency', 'randomness',
                                   'neurons', 'netstim', 'netcon', 'label'])


"""
Note to self: it would best best if PointNetwork would not actually
build the model until it's executed. That way we can run the entire
simulation in a separate thread (or multiple cores).
"""

class PointNetwork:
    """A Network of Leaky-Integrate-and-Fire (LIF) point processes.

    Examples
    --------
    >>> import navis.interfaces.neuron as nrn
    >>> N = nrn.PointNetwork()
    >>>
    """

    def __init__(self):
        self._neurons = []
        self._neurons_dict = {}
        self._edges = []
        self._stimuli = []
        self._ids = []
        self._labels = []
        self.idx = NetworkIdIndexer(self)

    def __str__(self):
        return f'{type(self).__name__}<neurons={len(self)},edges={len(self.edges)}>'

    def __repr__(self):
        return self.__str__()

    def __contains__(self, id):
        return id in self._ids

    def __getitem__(self, ix):
        """Get point process with given ID."""
        return np.asarray(self.neurons)[ix]

    def __len__(self):
        return len(self._neurons)

    @property
    def edges(self):
        """Edges between nodes of the network.

        Returns
        -------
        list of tuples

                `[(source_ix, target_ix, weight, NetCon object), ...]`

        """
        return self._edges

    @property
    def neurons(self):
        """Neurons in the network.

        Returns
        -------
        list
                List of `PointNeurons`.

        """
        return self._neurons

    @property
    def ids(self):
        """IDs of neurons in the network."""
        return self._ids

    @property
    def labels(self):
        """Labels of neurons in the network."""
        return self._labels

    @classmethod
    def from_edge_list(cls, edges, model='IntFire1', source_col='source',
                       target_col='target', weight_col='weight', **props):
        """Generate network from edge list.

        Parameters
        ----------
        edges :         pd.DataFrame
                        Edge list. Must contain 'source', 'target' and 'weight'
                        columns.
        model :         "IntFire1" | "IntFire2" | "IntFire4"
                        The model to use for the integrate-and-fire point processes.
        source_col :    str
                        Name of the column with the source IDs.
        target_col :    str
                        Name of the column with the target IDs.
        weight_col :    str
                        Name of the column with the weights. The important thing
                        to note here is that weight is expected to be in the 0-1
                        range with 1 effectively guaranteeing that a presynaptic
                        spike triggers a postsynaptic spike.
        **props
                        Keyword arguments are passed through to `add_neurons`.
                        Use to set e.g. labels, threshold or additional
                        model parameters.

        """
        assert isinstance(edges, pd.DataFrame)
        miss = {source_col, target_col, weight_col} - set(edges.columns)
        if miss:
            raise ValueError(f'edge list is missing columns: {miss}')

        # Instantiate
        net = cls()

        # Add neurons
        ids = np.unique(edges[[source_col, target_col]].values)
        net.add_neurons(ids, model=model, **props)

        # Connect neurons
        for (s, t, w) in edges[[source_col, target_col, weight_col]].values:
            net.connect(s, t, w)

        return net

    def add_neurons(self, ids, model='IntFire1', labels=None,
                    skip_existing=False, **props):
        """Add neurons to network.

        Parameters
        ----------
        ids :           list-like
                        Iterable of IDs for which to create neurons.
        model :         "IntFire1" | "IntFire2" | "IntFire4"
                        The model to use for the integrate-and-fire point processes.
        labels :        str | list | dict, optional
                        Labels for neurons. If str will apply the same label to
                        all neurons. If list, must be same length as `ids`.
                        Dictionary must be ID -> label map.
        skip_existing : bool
                        If True, will skip existing IDs.
        **props
                        Additional parameters used when initializing the point
                        processes. Depends on which model type you are using:
                        e.g. `refrac` and `tau` for "IntFire1".

        """
        assert model in ["IntFire1", "IntFire2", "IntFire4"]
        model = getattr(neuron.h, model)

        ids = utils.make_iterable(ids)

        if isinstance(labels, dict):
            labels = [labels.get(i, 'NA') for i in ids]
        elif labels:
            labels = utils.make_iterable(labels)
        else:
            labels = ids

        if len(labels) != len(ids):
            raise ValueError('Must provide a label for each neuron')

        for i, l in zip(ids, labels):
            if i in self:
                if skip_existing:
                    continue
                else:
                    raise ValueError(f'Neuron with id {i} already exists. '
                                     'Try using `skip_existing=True`.')

            # Create neuron
            n = PointNeuron(id=i, model=model, label=l, **props)

            self._neurons.append(n)
            self._ids.append(i)
            self._labels.append(l)
            self._neurons_dict[i] = n

    def add_background_noise(self, ids, frequency, randomness=.5,
                             independent=True):
        """Add background noise to given neurons.

        Parameters
        ----------
        ids :           hashable  | iterable
                        IDs of neurons to add noise to.
        frequency :     int | float | iterable
                        Frequency [Hz] of background noise. If iterable must
                        match length of `ids`.
        randomness :    float [0-1]
                        Randomness of spike timings.
        independent :   bool
                        If True (default), each neuron will get its own
                        independent noise stimulus.

        """
        self.add_stimulus(ids=ids, start=0, stop=9999999999,
                          frequency=frequency, randomness=randomness,
                          independent=independent)

    def add_stimulus(self, ids, start, frequency, stop=None, duration=None,
                     randomness=.5, independent=True, label=None, weight=1):
        """Add stimulus to given neurons.

        Important
        ---------
        Stimuli are implemented via a NetStim object which provides the
        specified stimulation (i.e. start, stop, frequency). This NetStim is
        then connected to the neuron(s) via a NetCon. The response of the
        neuron to the stimulus depends heavily on the `weight` of that
        connection: too low and you won't elicit any spikes, too high and you
        will produce higher frequencies than expected. The "correct" weight
        depends on the model & parameters you use for your point processes, and
        I highly recommend you check if you get the expected stimulation.

        Parameters
        ----------
        ids :           int | iterable
                        IDs of neurons to add stimulus to.
        start :         int
                        Start time [ms] for the stimulus. Note that
                        the exact start and stop time will vary depending on
                        `randomness`.
        stop/duration : int
                        Either stop time or duration of stimulus [ms]. Must
                        provide one or the other but not both.
        frequency :     int | iterable
                        Frequency [Hz] of background noise . If iterable must
                        match length of `ids`. Values <= 0 are silently skipped.
        randomness :    float [0-1]
                        Randomness of spike timings.
        independent :   bool
                        If True (default), each neuron will get its own
                        independent stimulus.
        label :         str, optional
                        A label to identify the stimulus.
        weight :        float
                        Weight for the connection between the stimulator and
                        the neuron. This really should be 1 to make sure each
                        spike in the stimulus elicits a spike in the target.

        """
        ids = utils.make_iterable(ids)
        if not utils.is_iterable(frequency):
            frequency = [frequency] * len(ids)
        elif not independent:
            raise ValueError('Stimuli/noises must be independent when '
                             'providing individual frequencies')

        if len(frequency) != len(ids):
            raise ValueError('Must provide either a single frequency for all '
                             'neurons or a frequency for each neuron.')

        if (duration and stop) or (not duration and not stop):
            raise ValueError('Must provide either duration or stop (but not both).')

        if stop:
            duration = stop - start
        else:
            stop = start + duration

        if duration <= 0:
            raise ValueError(f'Duration must be greater than zero.')

        if not independent:
            ns = neuron.h.NetStim()
            ns.interval = 1000 / frequency[0]
            ns.noise = randomness
            ns.number = int(duration / 1000 * frequency[0])
            ns.start = start

        for i, f in zip(ids, frequency):
            # Skip frequencies lower than 0
            if f <= 0:
                continue

            interval = 1000 / f
            proc = self.idx[i].process

            if independent:
                ns = neuron.h.NetStim()
                ns.interval = interval
                ns.noise = randomness
                ns.number = int(duration / 1000 * f)
                ns.start = start

            nc = neuron.h.NetCon(ns, proc)
            nc.weight[0] = weight
            nc.delay = 0

            self._stimuli.append(Stimulus(start, stop, f, randomness, i, ns, nc, label))

    def clear_stimuli(self):
        """Clear stimuli."""
        self._stimuli = []

    def connect(self, source, target, weight, delay=5):
        """Connect two neurons.

        Parameters
        ----------
        source :    int | str
                    ID of the source.
        target :    int | str
                    ID of the target
        weight :    float
                    Weight of the edge. The important thing to note here is that
                    the weight is expected to be in the 0-1 range with 1
                    effectively guaranteeing that a presynaptic spike triggers
                    a postsynaptic spike.
        delay :     int
                    Delay in ms between a pre- and a postsynaptic spike.

        """
        # Get the point processes corresponding to source and target
        pre = self.idx[source]
        post = self.idx[target]

        # Connect
        nc = neuron.h.NetCon(pre.process, post.process)
        nc.weight[0] = weight
        nc.delay = delay

        # Keep track
        self._edges.append([source, target, weight, nc])

    def plot_raster(self, subset=None, group=False, stimuli=True, ax=None, label=False,
                    backend='auto', **kwargs):
        """Raster plot of spike timings.

        Parameters
        ----------
        subset :        list-like
                        Subset of IDs to plot. You can also use this to
                        determine the order of appearance.
        ax :            matplotlib axis |  plotly figure, optional
                        Axis/figure to plot onto.
        label :         bool
                        Whether to label individual neurons.
        backend :       "auto" | "plotly" | "matplotlib"
                        Which backend to use. If "auto" will use plotly in
                        Jupyter environments and matplotlib elsewhere.

        """
        if not isinstance(subset, type(None)):
            ids = utils.make_iterable(subset)
            if not len(ids):
                raise ValueError('`ids` must not be empty')
        else:
            ids = self._ids

        # Collect spike timings
        x = []
        y = []
        i = 0
        for id in ids:
            pp = self.idx[id]
            x += list(pp.spk_timings)
            y += [i] * len(pp.spk_timings)
            i += 1

        if not x:
            raise ValueError('No spikes detected.')

        if label:
            ld = dict(zip(self._ids, self._labels))
            labels = [ld[i] for i in ids]
        else:
            labels = None

        # Turn into lines
        x = np.vstack((x, x, [None] * len(x))).T.flatten()
        y = np.array(y)
        y = np.vstack((y, y + .9, [None] * len(y))).T.flatten()

        if backend == 'auto':
            if utils.is_jupyter():
                backend = 'plotly'
            else:
                backend = 'matplotlib'

        if backend == 'plotly':
            return _plot_raster_plotly(x, y, ids, fig=ax, labels=labels, **kwargs)
        elif backend == 'matplotlib':
            ax = _plot_raster_mpl(x, y, ids, ax=ax, labels=labels,
                                  stimuli=self._stimuli if stimuli else None,
                                  **kwargs)
            ax.set_xlim(0, neuron.h.t)
            return ax
        else:
            raise ValueError(f'Unknown backend "{backend}"')

    def plot_traces(self, bin_size=100, subset=None, rolling_window=None,
                    group=False, stimuli=True, ax=None, backend='auto', **kwargs):
        """Plot mean firing rate.

        Parameters
        ----------
        bin_size :          int
                            Size [ms] of the bins over which to average spike
                            frequency.
        subset :            list-like
                            Subset of IDs to plot. You can also use this to
                            determine the order of appearance.
        rolling_window :    int
                            Average firing rates over a given rolling window.
        group :             bool | array
                            Set to True to group traces by label, showing mean
                            firing rate and standard error as envelope. Pass
                            an array of labels for each neuron to group by
                            arbitrary labels.
        ax :                matplotlib axis | plotly figure, optional
                            Axis/figure to plot onto.
        stimuli :           bool
                            Whether to plot stimuli.
        backend :           "auto" | "plotly" | "matplotlib"
                            Which backend to use. If "auto" will use plotly in
                            Jupyter environments and matplotlib elsewhere.

        """
        if not isinstance(subset, type(None)):
            ids = utils.make_iterable(subset)
            if not len(ids):
                raise ValueError('`ids` must not be empty')
        else:
            ids = self._ids

        # Collect spike frequencies
        spks = self.get_spike_counts(bin_size=bin_size, subset=subset,
                                     rolling_window=rolling_window)
        freq = spks * 1000 / bin_size

        if self.labels:
            ld = dict(zip(self._ids, self._labels))
            labels = [f'{i} ({ld[i]})' for i in ids]
        else:
            labels = None

        if backend == 'auto':
            if utils.is_jupyter():
                backend = 'plotly'
            else:
                backend = 'matplotlib'

        if isinstance(group, bool) and group:
            sem = freq.groupby(dict(zip(self._ids, self._labels))).sem()
            freq = freq.groupby(dict(zip(self._ids, self._labels))).mean()
            labels = freq.index.values.tolist()
        elif not isinstance(group, bool):
            sem = freq.groupby(group).sem()
            freq = freq.groupby(group).mean()
            labels = freq.index.values.tolist()
        else:
            sem = None

        if backend == 'plotly':
            return _plot_traces_plotly(freq, fig=ax, labels=labels,
                                       stimuli=self._stimuli if stimuli else None,
                                       env=sem, **kwargs)
        elif backend == 'matplotlib':
            return _plot_traces_mpl(freq, ax=ax, env=sem,
                                    stimuli=self._stimuli if stimuli else None,
                                    **kwargs)
        else:
            raise ValueError(f'Unknown backend "{backend}"')

    def set_labels(self, labels):
        """Set labels for neurons.

        Parameters
        ----------
        labels :    dict | list-like
                    If list, must provide a label for every neuron.

        """
        if isinstance(labels, dict):
            for i, n in enumerate(self.neurons):
                n.label = self._labels[i] = labels.get(n.id, n.id)
        elif utils.is_iterable(labels):
            if len(labels) != len(self.neurons):
                raise ValueError(f'Got {len(labels)} labels for {len(self)} neurons.')
            for i, n in enumerate(self.neurons):
                n.label = self._labels[i] = labels[i]
        else:
            raise TypeError(f'`labels` must be dict or list-like, got "{type(labels)}"')

    def run_simulation(self, duration=25 * ms, v_init=-65 * mV):
        """Run the simulation."""

        # This resets the entire model space not just this neuron!
        neuron.h.finitialize(v_init)
        neuron.h.continuerun(duration)

    def get_spike_counts(self, bin_size=50, subset=None, rolling_window=None,
                         group=False):
        """Get matrix of spike counts.

        Parameters
        ----------
        bin_size :          int | None
                            Size [ms] of the bins over which to count spikes.
                            If None, will simply return total counts.
        rolling_window :    int, optional
                            Average spike counts in a rolling window.
        group :             bool
                            If True, will return the spike counts per unique
                            label.

        Returns
        -------
        pd.DataFrame

        """
        if not isinstance(subset, type(None)):
            ids = utils.make_iterable(subset)
        else:
            ids = self._ids

        end_time = neuron.h.t

        if not end_time:
            raise ValueError('Looks like simulation has not yet been run.')

        if not bin_size:
            bin_size = end_time

        bins = np.arange(0, end_time + bin_size, bin_size)
        counts = np.zeros((len(ids), len(bins) - 1))
        # Collect spike counts
        for i, id in enumerate(ids):
            pp = self.idx[id]
            timings = list(pp.spk_timings)
            if timings:
                hist, _ = np.histogram(timings, bins)
                counts[i, :] = hist

        counts = pd.DataFrame(counts, index=ids, columns=bins[1:])

        if group:
            if not self._labels:
                raise ValueError('Unable to group: Network has no labels.')
            counts = counts.groupby(counts.index.map(dict(zip(self.ids, self._labels)))).sum()

        if rolling_window:
            avg = sliding_window_view(counts, rolling_window, axis=1).mean(axis=2)
            counts.iloc[:, :-(rolling_window - 1)] = avg

        return counts


class PointNeuron:
    def __init__(self, id, model, label=None, **props):
        self.process = model()
        self.id = id
        self.label = label
        for p, v in props.items():
            setattr(self.process, p, v)

        self.record_spikes()

    def __str__(self):
        return f'{type(self).__name__}<id={self.id},label={self.label}>'

    def __repr__(self):
        return self.__str__()

    def record_spikes(self, threshold=10):
        """Set up spike recording for this neuron."""
        self.spk_timings = neuron.h.Vector()
        self.sp_det = neuron.h.NetCon(self.process, None)
        self.sp_det.record(self.spk_timings)

    def record_state(self):
        """Set up recording of state for this neuron."""
        self.state = neuron.h.Vector()
        self.state.record(self.process._ref_m)


class NetworkIdIndexer:
    def __init__(self, network):
        self.network = network

    def __getitem__(self, id):
        neurons = self.network._neurons_dict
        if utils.is_iterable(id):
            return [neurons[i] for i in id]
        else:
            return neurons[id]


def _plot_raster_mpl(x, y, ids, ax=None, labels=None, stimuli=None, **kwargs):
    if not ax:
        fig, ax = plt.subplots(figsize=kwargs.pop('figsize', (12, min(20, len(ids)))))

    DEFAULTS = dict(alpha=.9, rasterized=False, lw=1)
    DEFAULTS.update(kwargs)

    ax.plot(x, y, **DEFAULTS)
    ax.set_ylim(-.5, len(ids) + .5)
    ax.set_xlabel('time [ms]')

    if not isinstance(labels, type(None)):
        ax.set_yticks(np.arange(0, len(ids), 1) + .5)
        ax.set_yticklabels(labels)
    else:
        ax.set_yticks([])
        ax.set_yticklabels([])

    if stimuli:
        y = len(ids) + max(1, len(ids) / 100)
        stimuli = np.unique([(s.start, s.stop) for s in stimuli], axis=0)
        for st in stimuli:
            # Skip background noise
            if st[1] >= 999_999_999:
                continue
            ax.plot([st[0], st[1]], [y, y], lw=4,
                    color=kwargs.get('color', (.5, .5, .5)))

        ax.set_ylim(top=y + 1)

    return ax


def _plot_traces_mpl(freq, ax=None, show=True, env=None, stimuli=None, **kwargs):
    if not ax:
        fig, ax = plt.subplots(figsize=kwargs.pop('figsize', (12, 7)))

    if 'color' in kwargs:
        c = kwargs.pop('color')
        if utils.is_iterable(c) and len(c) == freq.shape[0]:
            colors = np.array([mcl.to_rgb(c) for c in c])
            if colors.max() > 1:
                colors /= 255
        else:
            c = mcl.to_rgb(c)
            if max(c) > 1:
                c = (np.array(c) / 255).astype(int)
            colors = [c] * freq.shape[0]
    else:
        colors = sns.color_palette('tab20', freq.shape[0])

    segs = np.zeros((freq.shape[0], freq.shape[1], 2))
    segs[:, :, 0] = freq.columns.values
    segs[:, :, 1] = freq.values

    DEFAULTS = dict(alpha=.9, rasterized=False, lw=1)
    DEFAULTS.update(kwargs)

    lc = LineCollection(segs, **DEFAULTS, colors=colors)
    ax.add_collection(lc)

    if not isinstance(env, type(None)):
        for i, s in enumerate(env.index.values):
            # If no envelope (single neuron)
            if not env.loc[s].any():
                continue

            y1 = freq.loc[s].values + env.loc[s].values
            y2 = freq.loc[s].values - env.loc[s].values
            ax.fill_between(env.columns.values, y1, y2,
                            facecolor=colors[i],
                            alpha=0.5)

    if stimuli:
        y = freq.max().max() + 20
        stimuli = np.unique([(s.start, s.stop) for s in stimuli], axis=0)
        for st in stimuli:
            # Skip background noise
            if st[1] >= 999_999_999:
                continue
            ax.plot([st[0], st[1]], [y, y], lw=4,
                    color=kwargs.get('color', (.5, .5, .5)))

    ax.set_xlabel('time [ms]')
    ax.set_ylabel('spike rate [Hz]')

    ax.autoscale()

    return ax


def _plot_raster_plotly(x, y, ids, fig=None, labels=None, show=True, **kwargs):
    if not fig:
        fig = go.Figure()

    if labels:
        # Turn into one label for each line
        labels = [labels[i] for i in y[::3]]
        ht = np.vstack((labels, labels, [None] * len(labels))).T.flatten().tolist()
    else:
        ht = None

    fig.add_trace(go.Scattergl(x=x,
                               y=y,
                               mode='lines',
                               hovertext=ht,
                               hoverinfo='text' if labels else 'x+y',
                               line=dict(color=kwargs.get('color', 'black'),
                                         width=kwargs.get('width', 2.5))))

    fig.update_layout(paper_bgcolor='rgba(0,0,0,0)',
                      height=kwargs.get('height', 1000),
                      xaxis_title="time [ms]",
                      plot_bgcolor='rgba(0,0,0,0)')

    if show:
        fig.show()

    return fig


def _plot_traces_plotly(freq, fig=None, labels=None, show=True, stimuli=None,
                        env=None, **kwargs):
    if not fig:
        fig = go.Figure()

    if 'color' in kwargs:
        c = kwargs['color']
        if utils.is_iterable(c) and len(c) == freq.shape[0]:
            colors = np.array([mcl.to_rgb(c) for c in c])
            if colors.max() <= 1 and colors.max() != 0:
                colors *= 255
        else:
            c = mcl.to_rgb(c)
            if max(c) <= 1 and max(c) != 0:
                c = (np.array(c) * 255).astype(int)
            colors = [c] * freq.shape[0]
    else:
        colors = sns.color_palette('tab20', freq.shape[0])
        colors = np.array(colors) * 255

    alpha = kwargs.get('alpha', .5)

    x = freq.columns.values
    for i in range(freq.shape[0]):
        if not any(freq.iloc[i].values > 0):
            continue
        color = colors[i]
        color_str = f'rgba({color[0]},{color[1]},{color[2]},{alpha})'
        fill_color_str = f'rgba({color[0]},{color[1]},{color[2]},{.1})'

        fig.add_trace(go.Scattergl(x=x,
                                   y=freq.iloc[i].values,
                                   mode='lines',
                                   hovertext=labels[i] if labels else None,
                                   hoverinfo='text' if labels else 'x+y',
                                   name=labels[i] if labels else freq.index[i],
                                   legendgroup=labels[i] if labels else freq.index[i],
                                   line=dict(color=color_str,
                                             width=kwargs.get('width', 1.5))))

        if not isinstance(env, type(None)):
            y = (freq.iloc[i].values + env.iloc[i].values).tolist()
            y += (freq.iloc[i].values - env.iloc[i].values).tolist()[::-1]
            fig.add_trace(go.Scattergl(x=x.tolist() + x.tolist()[::-1],
                                       y=y,
                                       fill='toself',
                                       fillcolor=fill_color_str,
                                       name=labels[i] if labels else freq.index[i],
                                       showlegend=False,
                                       hoverinfo='skip',
                                       legendgroup=labels[i] if labels else freq.index[i],
                                       line=dict(color="rgba(255,255,255,0)",
                                                 width=kwargs.get('width', 1.5))))

    if stimuli:
        timings = [(st.start, st.stop) for st in stimuli]
        to_plot = np.unique(timings, axis=0, return_index=True)[1]
        y = freq.max().max() + 20
        for i, ix in enumerate(to_plot):
            st = stimuli[ix]
            # Skip background noise
            if st.stop >= 999_999_999:
                continue
            fig.add_trace(go.Scattergl(x=[st.start, st.stop], y=[y+i, y+i],
                                       mode='lines',
                                       line=dict(color='rgb(155,155,155)',
                                                 width=4),
                                       name=st.label,
                                       hovertext=st.label,
                                       hoverinfo='text',
                                       showlegend=False))

    fig.update_layout(paper_bgcolor='rgba(0,0,0,0)',
                      height=kwargs.get('height', 500),
                      xaxis_title="time [ms]",
                      yaxis_title="spike rate [Hz]",
                      yaxis_gridwidth=.25,
                      xaxis_gridwidth=.25,
                      xaxis_showgrid=False,
                      yaxis_showgrid=False,
                      plot_bgcolor='rgba(0,0,0,0)')

    if show:
        fig.show()

    return fig



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import networkx as nx
import numpy as np

from .. import core


__all__ = ['health_check', 'merge_duplicate_nodes']


def health_check(x: 'core.NeuronObject', verbose: bool = True) -> None:
    """Run a health check on TreeNeurons and flag potential issues.

    Parameters
    ----------
    x :         TreeNeuron | NeuronList
                Neuron(s) whose nodes to classify nodes.
    verbose :   bool
                If True, will print errors in addition to returning them.

    Returns
    -------
    list of issues or None

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> navis.health_check(n)
    Neuron 1734350788 seems perfectly fine.

    """
    if isinstance(x, core.NeuronList):
        for n in x:
            _ = health_check(x)
        return
    elif not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Excepted TreeNeuron/List, got "{type(x)}"')

    issues = []

    # Check if neuron is not a tree
    if not x.is_tree:
        issues.append('is not a tree (networkx.is_forest)')
    # See if there are any cycles
    if x.cycles:
        issues.append(f'has cycles (networkx.find_cycles): {str(x.cycles)}')
    # See if any node has more than one parent
    od = [n[0] for n in x.graph.out_degree if n[1] > 1]
    if od:
        issues.append(f'has nodes with multiple parents (graph.out_degree): {", ".join(od)}')

    locs, counts = np.unique(x.nodes[['x', 'y', 'z']].values,
                             axis=0,
                             return_counts=True)
    dupl = counts > 1
    if any(dupl):
        issues.append(f'has {sum(dupl)} node positions that are occupied by multiple nodes')

    if verbose:
        if issues:
            print(f'Neuron {str(x.id)} has issues:')
            for i in issues:
                print(f' - {i}')
        else:
            print(f'Neuron {str(x.id)} seems perfectly fine.')

    return issues if issues else None


def merge_duplicate_nodes(x, round=False, inplace=False):
    """Merge nodes the occupy the exact same position in space.

    Note that this might produce connections where there previously weren't
    any!

    Parameters
    ----------
    x :         TreeNeuron | NeuronList
                Neuron(s) to fix.
    round :     int, optional
                If provided will round node locations to given decimals. This
                can be useful if the positions are floats and not `exactly` the
                the same.
    inplace :   bool
                If True, perform operation on neuron inplace.

    Returns
    -------
    TreeNeuron
                Fixed neuron. Only if `inplace=False`.

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> n.nodes.loc[1, ['x', 'y' ,'z']] = n.nodes.loc[0, ['x', 'y' ,'z']]
    >>> fx = navis.graph.clinic.merge_duplicate_nodes(n)
    >>> n.n_nodes, fx.n_nodes
    (4465, 4464)

    """
    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()

        for n in x:
            _ = merge_duplicate_nodes(n, round=round, inplace=True)

        if not inplace:
            return x
        return

    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Expected TreeNeuron, got "{type(x)}"')

    if not inplace:
        x = x.copy()

    # Figure out which nodes are duplicated
    if round:
        dupl = x.nodes[['x', 'y', 'z']].round(round).duplicated(keep=False)
    else:
        dupl = x.nodes[['x', 'y', 'z']].duplicated(keep=False)

    if dupl.sum():
        # Operate on the edge list
        edges = x.nodes[['node_id', 'parent_id']].values.copy()

        # Go over each non-unique location
        ids = x.nodes.loc[dupl].groupby(['x', 'y', 'z']).node_id.apply(list)
        for i in ids:
            # Keep the first node and collapse all others into it
            edges[np.isin(edges[:, 0], i[1:]), 0] = i[0]
            edges[np.isin(edges[:, 1], i[1:]), 1] = i[0]

        # Drop self-loops
        edges = edges[edges[:, 0] != edges[:, 1]]

        # Make sure we don't have a->b and b<-a edges
        edges = np.unique(np.sort(edges, axis=1), axis=0)

        G = nx.Graph()

        # Get nodes but drop ""-1"
        nodes = edges.flatten()
        nodes = nodes[nodes >= 0]

        # Add nodes
        G.add_nodes_from(nodes)

        # Drop edges that point away from root (e.g. (1, -1))
        # Don't do this before because we would loose isolated nodes otherwise
        edges = edges[edges.min(axis=1) >= 0]

        # Add edges
        G.add_edges_from([(e[0], e[1]) for e in edges])

        # First remove cycles
        while True:
            try:
                # Find cycle
                cycle = nx.find_cycle(G)
            except nx.exception.NetworkXNoCycle:
                break
            except BaseException:
                raise

            # Sort by degree
            cycle = sorted(cycle, key=lambda x: G.degree[x[0]])

            # Remove the edge with the lowest degree
            G.remove_edge(cycle[0][0], cycle[0][1])

        # Now make sure this is a DAG, i.e. that all edges point in the same direction
        new_edges = []
        for c in nx.connected_components(G.to_undirected()):
            sg = nx.subgraph(G, c)

            # Try picking a node that was root in the original neuron
            is_present = np.isin(x.root, sg.nodes)
            if any(is_present):
                r = x.root[is_present][0]
            else:
                r = list(sg.nodes)[0]

            # Generate parent->child dictionary by graph traversal
            this_lop = nx.predecessor(sg, r)

            # Note that we assign -1 as root's parent
            new_edges += [(k, v[0]) for k, v in this_lop.items() if v]

        # We need a directed Graph for this as otherwise the child -> parent
        # order in the edges might get lost
        G2 = nx.DiGraph()
        G2.add_nodes_from(G.nodes)
        G2.add_edges_from(new_edges)

        # Generate list of parents
        new_edges = np.array(G2.edges)
        new_parents = dict(zip(new_edges[:, 0], new_edges[:, 1]))

        # Drop nodes that aren't present anymore
        x._nodes = x._nodes.loc[x._nodes.node_id.isin(new_edges.flatten())].copy()

        # Rewire kept nodes
        x.nodes['parent_id'] = x.nodes.node_id.map(lambda x: new_parents.get(x, -1))

        # Reset temporary attributes
        x._clear_temp_attr()

    if not inplace:
        return x



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
"""Module contains network models."""

import multiprocessing as mp
import numpy as np
import pandas as pd
import warnings

from typing import Iterable, Union, Optional, Callable

from .. import config

# Set up logging
logger = config.get_logger(__name__)

__all__ = ['BayesianTraversalModel', 'TraversalModel', 'linear_activation_p',
           'random_linear_activation_function']


class BaseNetworkModel:
    """Base model for network simulations."""

    def __init__(self, edges: pd.DataFrame, source: str, target: str):
        """Initialize model."""
        assert isinstance(edges, pd.DataFrame), f'edges must be pandas DataFrame, got "{type(edges)}"'
        assert source in edges.columns, f'edges DataFrame must contain "{source}" column'
        assert target in edges.columns, f'edges DataFrame must contain "{target}" column'
        self.edges = edges
        self.source = source
        self.target = target

        if (self.edges.dtypes[source] == object) or (self.edges.dtypes[target] == object):
            logger.warning('Looks like sources and/or targets in your edge list '
                           'might be strings? This can massively slow down '
                           'computations. If at all possible try to use numeric IDs.')

    @property
    def n_nodes(self) -> int:
        """Return unique nodes in network."""
        return np.unique(self.edges[[self.source, self.target]].values.flatten()).shape[0]

    @property
    def has_results(self) -> bool:
        """Check if model has results."""
        if isinstance(getattr(self, 'results', None), pd.DataFrame):
            return True
        return False

    def run_parallel(self,
                     n_cores: int = 5,
                     iterations: int = 100,
                     **kwargs: dict) -> None:
        """Run model using parallel processes."""
        # Note that we initialize each process by making "edges" a global argument
        with mp.Pool(processes=n_cores,
                     initializer=self.initializer) as pool:

            # Each process will take about the same amount of time
            # So we want each process to take a single batch of iterations/n_cores runs
            kwargs['iterations'] = int(iterations/n_cores)
            calls = [{**kwargs, **{'position': i}} for i in range(int(n_cores))]

            # Generate processes - note the use of chunksize 1 because we have
            # already chunked the iterations such that each worker process
            # runs exactly once
            p = pool.imap_unordered(self._worker_wrapper, calls, chunksize=1)

            # Wait for processes to complete
            res = list(p)

        # Combine results
        self.results = pd.concat(res, axis=0)
        self.iterations = iterations

    def _worker_wrapper(self, kwargs: dict):
        return self.run(**kwargs)

    def initializer(self):
        pass

    def run(self):
        pass


class TraversalModel(BaseNetworkModel):
    """Model for traversing a network starting with given seed nodes.

    What this does:

      1. Grab all already visited nodes (starting with `seeds` in step 1)
      2. Find all downstream nodes of these
      3. Probabilistically traverse based on the weight of the connecting edges
      4. Add those newly visited nodes to the pool & repeat from beginning
      5. Stop when every (connected) neuron was visited or we reached `max_steps`

    Parameters
    ----------
    edges :             pandas.DataFrame
                        DataFrame representing an edge list. Must minimally have
                        a `source` and `target` column.
    seeds :             iterable
                        Seed nodes for traversal. Nodes that aren't found in
                        `edges['source']` will be (silently) removed.
    weights :           str, optional
                        Name of a column in `edges` used as weights. If not
                        provided, all edges will be given a weight of 1. If using
                        the default activation function the weights need to be
                        between 0 and 1.
    max_steps :         int
                        Limits the number of steps for each iteration.
    traversal_func :    callable, optional
                        Function that determines whether a given edge will be
                        traversed or not in a given step. Must take numpy array
                        (N, 1) of edge weights and return an array with
                        True/False of equal size. Defaults to
                        [`navis.models.network_models.random_linear_activation_function`][]
                        which will linearly scale probability of traversal
                        from 0 to 100% between edges weights 0 to 0.3.

    Examples
    --------
    >>> from navis.models import TraversalModel
    >>> import networkx as nx
    >>> import numpy as np
    >>> # Generate a random graph
    >>> G = nx.fast_gnp_random_graph(1000, .2, directed=True)
    >>> # Turn into edge list
    >>> edges = nx.to_pandas_edgelist(G)
    >>> # Add random edge weights
    >>> edges['weight'] = np.random.random(edges.shape[0])
    >>> # Initialize model
    >>> model = TraversalModel(edges, seeds=list(G.nodes)[:10])
    >>> # Run model on 2 cores
    >>> model.run_parallel(n_cores=2, iterations=100)
    >>> # Get a summary
    >>> model.summary.tail()                                    # doctest: +SKIP
          layer_min  layer_max  layer_mean  layer_median
    node
    995           2          2        2.00             2
    996           2          3        2.33             2
    997           2          2        2.00             2
    998           2          2        2.00             2
    999           2          2        2.00             2

    Above Graph was traversed quickly (3 steps max). Let's adjust the
    traversal function:

    >>> from navis.models import random_linear_activation_function
    >>> # Use a lower probability for activation
    >>> def my_act(x):
    ...     return random_linear_activation_function(x, max_w=10)
    >>> model = TraversalModel(edges, seeds=list(G.nodes)[:10],
    ...                        traversal_func=my_act)
    >>> res = model.run(iterations=100)
    >>> res.tail()                                              # doctest: +SKIP
          layer_min  layer_max  layer_mean  layer_median
    node
    995           2          4       3.210           3.0
    996           2          4       3.280           3.0
    997           2          4       3.260           3.0
    998           2          4       3.320           3.0
    999           2          4       3.195           3.0

    """

    def __init__(self,
                 edges: pd.DataFrame,
                 seeds: Iterable[Union[str, int]],
                 source: str = 'source',
                 target: str = 'target',
                 weights: Optional[str] = 'weight',
                 max_steps: int = 15,
                 traversal_func: Optional[Callable] = None):
        """Initialize model."""
        super().__init__(edges=edges, source=source, target=target)

        if not weights:
            edges['weight'] = 1
            weights = 'weight'

        assert weights in edges.columns, f'"{weights}" must be column in edge list'

        # Remove seeds that don't exist
        self.seeds = edges[edges[self.source].isin(seeds)][self.source].unique()

        if len(self.seeds) == 0:
            raise ValueError('None of the seeds where among edge list sources.')

        self.weights = weights
        self.max_steps = max_steps

        if isinstance(traversal_func, type(None)):
            self.traversal_func = random_linear_activation_function
        elif callable(traversal_func):
            self.traversal_func = traversal_func
        else:
            raise ValueError('`traversal_func` must be None or a callable')

    @property
    def summary(self) -> pd.DataFrame:
        """Per-node summary."""
        return getattr(self, '_summary', self.make_summary())

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = f'{self.__class__}: {self.edges.shape[0]} edges; {self.n_nodes}' \
            f' unique nodes; {len(self.seeds)} seeds;' \
            f' traversal_func {self.traversal_func}.'
        if self.has_results:
            s += f' Model ran with {self.iterations} iterations.'
        else:
            s += ' Model has not yet been run.'
        return s

    def make_summary(self) -> pd.DataFrame:
        """Generate summary."""
        if not self.has_results:
            logger.error('Must run simulation first.')

        summary = self.results.groupby('node',
                                       as_index=False).steps.agg(['min',
                                                                  'max',
                                                                  'mean',
                                                                  'median'])

        summary.rename({'min': 'layer_min',
                        'mean': 'layer_mean',
                        'max': 'layer_max',
                        'median': 'layer_median'}, axis=1, inplace=True)

        self._summary = summary
        return self._summary

    def run(self, iterations: int = 100, return_iterations=False, **kwargs) -> pd.DataFrame:
        """Run model (single process).

        Use `.run_parallel` to use parallel processes.

        """
        # For some reason this is required for progress bars in Jupyter to show
        print(' ', end='', flush=True)
        # For faster access, use the raw array
        # Note: we're splitting the columns in case we have different datatypes
        # (e.g. int64 for IDs and float for weights)
        sources = self.edges[self.source].values
        targets = self.edges[self.target].values
        weights = self.edges[self.weights].values

        # For some reason the progress bar does not show unless we have a print here
        all_enc_nodes = None
        for it in config.trange(1, iterations + 1,
                                disable=config.pbar_hide,
                                leave=config.pbar_leave,
                                position=kwargs.get('position', 0)):
            # Set seeds as encountered in step 1
            enc_nodes = self.seeds
            enc_steps = np.repeat(1, len(self.seeds))
            if return_iterations:
                enc_it = np.repeat(it, len(self.seeds))

            # Start with all edges
            this_weights = weights
            this_sources = sources
            this_targets = targets
            for i in range(2, self.max_steps + 1):
                # Which edges have their presynaptic node already traversed?
                pre_trav = np.isin(this_sources, enc_nodes)
                # Among those, which edges have the postsynaptic node traversed?
                post_trav = np.isin(this_targets[pre_trav], enc_nodes)

                # Combine conditions to find edges where the presynaptic node
                # has been traversed but not the postsynaptic node
                pre_not_post = np.where(pre_trav)[0][~post_trav]
                out_targets = this_targets[pre_not_post]
                out_weights = this_weights[pre_not_post]

                # Drop edges that have already been traversed - speeds up things
                pre_and_post = np.where(pre_trav)[0][post_trav]
                this_targets = np.delete(this_targets, pre_and_post, axis=0)
                this_sources = np.delete(this_sources, pre_and_post, axis=0)
                this_weights = np.delete(this_weights, pre_and_post, axis=0)

                # Stop if we traversed the entire (reachable) graph
                if out_targets.size == 0:
                    break

                # Edges traversed in this round
                trav = self.traversal_func(out_weights)
                if not trav.sum():
                    continue

                trav_targets = out_targets[trav]

                # Keep track
                new_trav = np.unique(trav_targets)

                # Store results
                enc_nodes = np.concatenate((enc_nodes, new_trav))
                enc_steps = np.concatenate((enc_steps, np.repeat(i, len(new_trav))))
                if return_iterations:
                    enc_it = np.concatenate((enc_it, np.repeat(it, len(new_trav))))

            # Save this round of traversal
            if all_enc_nodes is None:
                all_enc_nodes = enc_nodes
                all_enc_steps = enc_steps
                if return_iterations:
                    all_enc_it = enc_it
            else:
                all_enc_nodes = np.concatenate((all_enc_nodes, enc_nodes))
                all_enc_steps = np.concatenate((all_enc_steps, enc_steps))
                if return_iterations:
                    all_enc_it = np.concatenate((all_enc_it, enc_it))

        self.iterations = iterations

        # Combine results into DataFrame
        self.results = pd.DataFrame()
        self.results['steps'] = all_enc_steps
        self.results['node'] = all_enc_nodes
        if return_iterations:
            self.results['iteration'] = all_enc_it

        return self.results


class BayesianTraversalModel(TraversalModel):
    """Model for traversing a network starting with given seed nodes.

    This model is a Bayes net version of
    [`navis.models.network_models.TraversalModel`][] that propagates
    traversal probabilities through the network and converges to a
    distribution of time of traversal for each node, rather than
    stochastically sampling.

    Unlike `TraversalModel`, this model should only be run once.
    Note alse that `traversal_func` should be a function returing
    probabilities of traversal, rather than a random boolean of traversal.

    Parameters
    ----------
    edges :             pandas.DataFrame
                        DataFrame representing an edge list. Must minimally have
                        a `source` and `target` column.
    seeds :             iterable
                        Seed nodes for traversal. Nodes that aren't found in
                        `edges['source']` will be (silently) removed.
    weights :           str, optional
                        Name of a column in `edges` used as weights. If not
                        provided, all edges will be given a weight of 1. If using
                        the default activation function the weights need to be
                        between 0 and 1.
    max_steps :         int
                        Limits the number of steps for each iteration.
    traversal_func :    callable, optional
                        Function returning probability whether a given edge will be
                        traversed or not in a given step. Must take numpy array
                        (N, 1) of edge weights and return an array with
                        probabilities of equal size. Defaults to
                        [`navis.models.network_models.linear_activation_p`][]
                        which will linearly scale probability of traversal
                        from 0 to 100% between edges weights 0 to 0.3.

    Examples
    --------
    >>> from navis.models import BayesianTraversalModel
    >>> import networkx as nx
    >>> import numpy as np
    >>> # Generate a random graph
    >>> G = nx.fast_gnp_random_graph(1000, .2, directed=True)
    >>> # Turn into edge list
    >>> edges = nx.to_pandas_edgelist(G)
    >>> # Add random edge weights
    >>> edges['weight'] = np.random.random(edges.shape[0])
    >>> # Initialize model
    >>> model = BayesianTraversalModel(edges, seeds=list(G.nodes)[:10])
    >>> # Run model
    >>> res = model.run()
    >>> # Get a summary
    >>> model.summary.tail()                                    # doctest: +SKIP
          layer_min  layer_max  layer_mean  layer_median
    node
    995           2          2        2.00             2
    996           2          3        2.33             2
    997           2          2        2.00             2
    998           2          2        2.00             2
    999           2          2        2.00             2

    Above Graph was traversed quickly (3 steps max). Let's adjust the
    traversal function:

    >>> from navis.models import linear_activation_p
    >>> # Use a lower probability for activation
    >>> def my_act(x):
    ...     return linear_activation_p(x, max_w=10)
    >>> model = BayesianTraversalModel(edges, seeds=list(G.nodes)[:10],
    ...                                traversal_func=my_act)
    >>> res = model.run()
    >>> res.tail()                                              # doctest: +SKIP
          layer_min  layer_max  layer_mean  layer_median
    node
    995           2          4       3.210           3.0
    996           2          4       3.280           3.0
    997           2          4       3.260           3.0
    998           2          4       3.320           3.0
    999           2          4       3.195           3.0

    """

    def __init__(self,
                 *args,
                 traversal_func: Optional[Callable] = None,
                 **kwargs):
        """Initialize model."""
        super().__init__(*args, **kwargs)

        if isinstance(traversal_func, type(None)):
            self.traversal_func = linear_activation_p
        elif callable(traversal_func):
            self.traversal_func = traversal_func
        else:
            raise ValueError('`traversal_func` must be None or a callable')

    def make_summary(self) -> pd.DataFrame:
        """Generate summary."""
        if not self.has_results:
            logger.error('Must run simulation first.')

        cmfs = np.stack(self.results.cmf)
        layer_min = (cmfs != 0).argmax(axis=1)
        valid = np.any(cmfs != 0, axis=1)
        layer_max = (cmfs == 1.).argmax(axis=1)
        layer_max[~np.any(cmfs == 1., axis=1)] = cmfs.shape[1]
        layer_median = (cmfs >= .5).argmax(axis=1).astype(float)
        pmfs = np.diff(cmfs, axis=1, prepend=0.)
        layer_pmfs = pmfs * np.arange(pmfs.shape[1])
        layer_mean = np.sum(layer_pmfs, axis=1)

        summary = pd.DataFrame({
                'layer_min': layer_min + 1,
                'layer_max': layer_max + 1,
                'layer_mean': layer_mean + 1,
                'layer_median': layer_median + 1,
            }, index=self.results.node)

        # Discard nodes that are never activated
        summary = summary[valid]

        self._summary = summary
        return self._summary

    def run(self, **kwargs) -> pd.DataFrame:
        """Run model (single process)."""

        # For some reason this is required for progress bars in Jupyter to show
        print(' ', end='', flush=True)
        # For faster access, use the raw array
        edges = self.edges[[self.source, self.target, self.weights]].values
        id_type = self.edges[self.source].dtype
        # Transform weights to traversal probabilities
        edges[:, 2] = self.traversal_func(edges[:, 2])

        # Change node IDs into indices in [0, len(nodes))
        ids, edges_idx = np.unique(edges[:, :2], return_inverse=True)
        ids = ids.astype(id_type)
        edges_idx = edges_idx.reshape((edges.shape[0], 2))
        edges_idx = np.concatenate(
            (edges_idx, np.expand_dims(edges[:, 2], axis=1)),
            axis=1)

        cmfs = np.zeros((len(ids), self.max_steps), dtype=np.float64)
        seed_idx = np.searchsorted(ids, self.seeds)
        cmfs[seed_idx, :] = 1.
        changed = set(edges_idx[np.isin(edges_idx[:, 0], seed_idx), 1].astype(id_type))

        with config.tqdm(
                total=0,
                disable=config.pbar_hide,
                leave=config.pbar_leave,
                position=kwargs.get('position', 0)) as pbar:
            while len(changed):
                next_changed = []

                pbar.total += len(changed)
                pbar.refresh()

                for idx in changed:
                    cmf = cmfs[idx, :]
                    inbound = edges_idx[edges_idx[:, 1] == idx, :]
                    pre = inbound[:, 0].astype(np.int64)

                    # Traversal probability for each inbound edge at each time.
                    posteriors = cmfs[pre, :] * np.expand_dims(inbound[:, 2], axis=1)
                    # At each time, compute the probability that at least one inbound edge
                    # is traversed.
                    new_pmf = 1 - np.prod(1 - posteriors, axis=0)
                    new_cmf = cmf.copy()
                    # Offset the time-cumulative probability by 1 to account for traversal iteration.
                    # Use maximum of previous CMF as it is monotonic and to include fixed seed traversal.
                    new_cmf[1:] = np.maximum(cmf[1:], 1 - np.cumprod(1 - new_pmf[:-1]))
                    np.clip(new_cmf, 0., 1., out=new_cmf)

                    if np.allclose(cmf, new_cmf):
                        continue

                    cmfs[idx, :] = new_cmf

                    # Notify downstream nodes that they have changed next iteration.
                    post_idx = edges_idx[edges_idx[:, 0] == idx, 1].astype(id_type)
                    next_changed.extend(list(post_idx))

                pbar.update(len(changed))
                changed = set(next_changed)

        self.iterations = 1
        self.results = pd.DataFrame({'node': ids, 'cmf': list(cmfs)})
        return self.results

    def run_parallel(self, *args, **kwargs) -> None:
        warnings.warn(f"{self.__class__.__name__} should not be run in parallel. Falling back to run.")
        self.run(**kwargs)


def linear_activation_p(
    w: np.ndarray,
    min_w: float = 0,
    max_w: float = .3,
) -> np.ndarray:
    """Linear activation probability.

    Parameters
    ----------
    w :     np.ndarray
            (N, 1) array containing the edge weights.
    min_w : float
            Value of `w` at which probability of activation is 0%.
    max_w : float
            Value of `w` at which probability of activation is 100%.

    Returns
    -------
    np.ndarray
            Probability of activation for each edge.

    """
    return np.clip((w - min_w) / (max_w - min_w), 0., 1.)


def random_activation_function(
    w: np.ndarray,
    func: Callable,
) -> np.ndarray:
    """Sample random activation for edges given a weight to probability function.

    Parameters
    ----------
    w :     np.ndarray
            (N, 1) array containing the edge weights.
    func :  callable
            A function returning activation probabilities given a set of weights.

    Returns
    -------
    np.ndarray
            True or False values for each edge.

    """
    # Generate a random number between 0 and 1 for each connection
    r = np.random.rand(w.shape[0])

    # Normalize weights to probabilities
    w_norm = func(w)

    # Test active
    act = w_norm >= r

    return act


def random_linear_activation_function(w: np.ndarray,
                                      min_w: float = 0,
                                      max_w: float = .3) -> np.ndarray:
    """Random linear activation function.

    Parameters
    ----------
    w :     np.ndarray
            (N, 1) array containing the edge weights.
    min_w : float
            Value of `w` at which probability of activation is 0%.
    max_w : float
            Value of `w` at which probability of activation is 100%.

    Returns
    -------
    np.ndarray
            True or False values for each edge.

    """
    return random_activation_function(w, lambda w: linear_activation_p(w, min_w, max_w))



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""
A collection of tools to interface with R natverse libraries (e.g. nat,
nblast, elmr, rcatmaid). See https://github.com/natverse.
"""
import math
import numbers
import os
import sys

from datetime import datetime
from colorsys import hsv_to_rgb
from typing import Union, Optional, List
from typing_extensions import Literal
from scipy.spatial.distance import pdist

import pandas as pd
import numpy as np

import rpy2
import rpy2.robjects as robjects
from rpy2.robjects.packages import importr

from rpy2.robjects import pandas2ri, numpy2ri

from .. import core, plotting, config, utils

# Inconveniently, rpy2's version vector differs in the way it's constructed
# between 2.X `((2, 9, 4), '')` and 3.X (3, 3, 2)
rpy2_major_version = int(rpy2.__version__.split('.')[0])
if rpy2_major_version >= 3:
    from rpy2.robjects.conversion import localconverter


# Set up logging
logger = config.get_logger(__name__)


def try_importr(x):
    """Try importing R library. Log error on exception."""
    try:
        return importr(x)
    except BaseException:
        logger.warning(f'Failed to import R library "{x}"! Some functions might'
                       ' not work as expected. Please install from within R.')
        # Return dummy class
        return FailedImport(x)


def try_loadr(x):
    """Try finding robject. Will postpone exception until use time."""
    try:
        return robjects.r(x)
    except BaseException:
        # Return dummy class
        return FailedObject(x)


class FailedImport:
    """Dummy class for failed imports from R. Throws meaningful exceptions."""

    def __init__(self, name):
        self.name = name

    def raise_error(self):
        raise Exception(f'R library "{self.name}" could not be imported. '
                        'Please make sure it is properly installed.')

    def __call__(self):
        self.raise_error()

    def __getattr__(self, name):
        self.raise_error()


class FailedObject(FailedImport):
    def raise_error(self):
        raise Exception(f'R object "{self.name}" was not found.')


cl = try_loadr('class')
names = try_loadr('names')
attributes = try_loadr('attributes')

# Load the entire natverse
nat = try_importr("nat")
r_nblast = try_importr('nat.nblast')
nat_templatebrains = try_importr('nat.templatebrains')
flycircuit = try_importr('flycircuit')
rcatmaid = try_importr('catmaid')

# These are mainly relevant for exposing transforms
nat_flybrains = try_importr('nat.flybrains')
elmr = try_importr('elmr')
nat_jrcbrains = try_importr('nat.jrcbrains')


__all__ = sorted(['neuron2r', 'neuron2py', 'init_rcatmaid',
                  'data2py', 'NBLASTresults', 'nblast', 'nblast_allbyall',
                  'get_neuropil', 'xform_brain', 'mirror_brain'])

# Do not use this! It will convert stuff in unexpected ways. E.g.
# dps = robjects.r(f'read.neuronlistfh("{datadir}/dpscanon.rds")')
# will turn out as just an array

# numpy2ri.activate()
# pandas2ri.activate()


def init_rcatmaid(**kwargs):
    """Initialize the R Catmaid package.

    R package by Greg Jefferis: https://github.com/jefferis/rcatmaid

    Parameters
    ----------
    remote_instance :   CATMAID instance
                        From pymaid.CatmaidInstance(). This is used to
                        extract credentials. Overrides other credentials
                        provided!
    server :            str, optional
                        Use this to set server URL if no remote_instance is
                        provided
    http_user :         str, optional
                        Use this to set http user if no remote_instance is
                        provided
    http_password :     str, optional
                        Use this to set http password if no remote_instance
                        is provided
    api_token :         str, optional
                        Use this to set user token if no remote_instance is
                        provided

    Returns
    -------
    catmaid :           R library
                        R object representing the rcatmaid library

    Examples
    --------
    >>> from navis.interfaces import r
    >>> rcatmaid = r.init_rcatmaid(server='https://your.catmaid.server',
    ...                            authname='http_user',
    ...                            authpassword='http_pw',
    ...                            authtoken='Your token')
    >>> # You can now use rcatmaid functions. For example:
    >>> neuron = rcatmaid.read_neurons_catmaid(16)
    >>> neuron
    R object with classes: ('neuronlist', 'list') mapped to:
    [ListSexpVector]
    16: <class 'rpy2.rinterface.ListSexpVector'>
    <rpy2.rinterface.ListSexpVector object at 0x123d46708> [RTYPES.VECSXP]

    """
    remote_instance = kwargs.get('remote_instance', None)
    server = kwargs.get('server', None)
    authname = kwargs.get('http_user', None)
    authpassword = kwargs.get('http_password', None)
    authtoken = kwargs.get('api_token', None)

    if remote_instance is None:
        if 'remote_instance' in sys.modules:
            remote_instance = sys.modules['remote_instance']
        elif 'remote_instance' in globals():
            remote_instance = globals()['remote_instance']

    if remote_instance:
        server = remote_instance.server
        authname = remote_instance.http_user
        authpassword = remote_instance.http_password
        authtoken = remote_instance.api_token
    elif not remote_instance and None in (server, authname, authpassword, authtoken):
        raise ValueError('Unable to initialize connection: missing credentials')

    # Use remote_instance's credentials
    rcatmaid.server = server
    rcatmaid.authname = authname
    rcatmaid.authpassword = authpassword
    rcatmaid.token = authtoken

    # Create the connection
    con = rcatmaid.catmaid_connection(server=rcatmaid.server,
                                      authname=rcatmaid.authname,
                                      authpassword=rcatmaid.authpassword,
                                      token=rcatmaid.token)

    # Login
    rcatmaid.catmaid_login(con)

    logger.info('Rcatmaid successfully initiated.')

    return rcatmaid


def data2py(data, **kwargs):
    """Convert data from rcatmaid (e.g. `catmaidneuron` from
    `read.neuron.catmaid`) and converts into Python Data.

    Notes
    -----
    (1) Most R data comes as list (even if only 1 entry). This is preserved.
    (2) R lists with headers are converted to dictionaries
    (3) R DataFrames are converted to Pandas DataFrames
    (4) R nblast results are converted to Pandas DataFrames but only the top
        100 hits for which we have reverse scores!

    Parameters
    ----------
    data
        Any kind of R data. Can be nested (e.g. list of lists)!

    Returns
    -------
    Converted data or 'Not converted' if conversion failed.

    """
    if 'neuronlistfh' in cl(data):
        logger.error('On-demand neuronlist found. Conversion cancelled to '
                     'prevent loading large datasets in memory. Please use '
                     'r.neuron2py() and its "subset" parameter.')
        return None
    elif 'neuronlist' in cl(data):
        return neuron2py(data)
    elif 'neuron' in cl(data):
        return neuron2py(data)
    elif 'dotprops' in cl(data):
        return neuron2py(data)
    elif cl(data)[0] == 'integer':
        if not names(data):
            return [int(n) for n in data]
        else:
            return {n: int(data[i]) for i, n in enumerate(names(data))}
    elif cl(data)[0] == 'character':
        if not names(data):
            return [str(n) for n in data]
        else:
            return {n: str(data[i]) for i, n in enumerate(names(data))}
    elif cl(data)[0] == 'numeric':
        if not names(data):
            return [float(n) for n in data]
        else:
            return {n: float(data[i]) for i, n in enumerate(names(data))}
    elif cl(data)[0] == 'data.frame':
        if rpy2.__version_vector__[0] < 3:
            return pandas2ri.ri2py(data)
        else:
            with localconverter(robjects.default_converter + pandas2ri.converter):
                return robjects.conversion.rpy2py(data)
    elif cl(data)[0] == 'matrix':
        mat = np.array(data)
        df = pd.DataFrame(data=mat)
        if data.names:
            if data.names[1] != robjects.r('NULL'):
                df.columns = data.names[1]
            if data.names[0] != robjects.r('NULL'):
                df.index = data.names[0]
        return df
    elif 'list' in cl(data):
        # If this is just a list, return a list
        if not names(data):
            return [data2py(n) for n in data]
        # If this list has headers, return as dictionary
        else:
            return {n: data2py(data[i]) for i, n in enumerate(names(data))}
    elif cl(data)[0] == 'NULL':
        return None
    elif 'nblastfafb' in cl(data):
        fw_scores = {n: data[0][i] for i, n in enumerate(names(data[0]))}
        rev_scores = {n: data[1][i] for i, n in enumerate(names(data[1]))}
        mu_scores = {n: (fw_scores[n] + rev_scores[n]) / 2 for n in rev_scores}

        df = pd.DataFrame([[n, fw_scores[n], rev_scores[n], mu_scores[n]]
                           for n in rev_scores],
                          columns=['gene_name', 'forward_score',
                                   'reverse_score', 'mu_score']
                          )

        logger.info('Returning only nblast results. Neuron object is stored '
                    'in your original_data[2].')
        return df
    else:
        logger.debug(f'Unable to convert R data of type "{cl(data)}"')
        return 'Not converted'


def neuron2py(x) -> 'core.NeuronObject':
    """Convert R neuron objects (dotprops, neurons, etc) to navis Neuron/Lists.

    Parameters
    ----------
    x :                 R dotprops | neuron | catmaidneuron | neuronlist
                        Neuron to convert to is navis analog.

    Returns
    -------
    Neuron/NeuronList

    """
    if 'rpy2' not in str(type(x)):
        raise TypeError(f'This does not look like R object: "{type(x)}"')

    # If neuronlist
    if 'neuronlist' in cl(x):
        data = [neuron2py(n) for n in x]
        return core.NeuronList(data)
    # If dotprops
    elif 'dotprops' in cl(x):
        # Convert data to Python
        data = {n: data2py(d) for n, d in zip(x.names, x)}
        data['k'] = int(x.slots['k'][0])
        # Make and return Dotprops
        return core.Dotprops(**data)
    # If neuron
    elif 'neuron' in cl(x):
        do_not_use = ['nTrees', 'SegList', 'NumPoints', 'StartPoint', 'EndPoints',
                      'BranchPoints', 'NumSegs']

        # Convert data to Python
        data = {n: data2py(d) for n, d in zip(x.names, x)
                if n not in do_not_use and isinstance(n, str)}

        # Construct neuron from just the nodes
        n = core.TreeNeuron(data.pop('d'))

        # R give node width, not radius - let's fix that
        if 'radius' in n.nodes.columns:
            has_rad = n.nodes.radius > 0
            n.nodes.loc[has_rad, 'radius'] = n.nodes.loc[has_rad, 'radius'] / 2

        # If this is a CATMAID neuron, we assume it's in nanometers
        if 'catmaidneuron' in cl(x):
            n.units = 'nm'

        # Reuse Id
        if 'skid' in data:
            if utils.is_iterable(data['skid']):
                n.id = data['skid'][0]
            else:
                n.id = data['skid']

        # Try attaching other data
        for k, v in data.items():
            try:
                setattr(n, k, v)
            except BaseException:
                pass
        return n
    else:
        raise TypeError(f'Unable to convert object of class "{cl(x)}"')


def neuron2r(x: 'core.NeuronObject',
             unit_conversion: Union[bool, int, float] = None,
             add_metadata: bool = False):
    """Convert Neuron/List to corresponding R neuron/neuronlist object.

    Parameters
    ----------
    x :                 TreeNeuron | Dotprops | NeuronList
    unit_conversion :   bool | int | float, optional
                        If not `False` will multiply units by given factor.
    add_metadata :      bool, optional
                        If False, will use minimal data necessary to construct
                        the R neuron. If True, will add additional data
                        associated with TreeNeuron - this could impact
                        R functions. **Currently not functional!**

    Returns
    -------
    R neuron
        Either R neuron or neuronlist depending on input.

    """
    if isinstance(x, core.NeuronList):
        """
        The way neuronlist are constructed is a bit more complicated:
        They are essentially named lists {'neuronA': neuronobject, ...} BUT
        they also contain a dataframe as attribute ( attr('df') = df )
        This dataframe looks like this

                pid   skid     name
        skid1
        skid2

        In rpy2, attributes are assigned using the .slots['df'] function.
        """

        nlist = {str(n.id): neuron2r(n, unit_conversion=unit_conversion) for
                 n in config.tqdm(x, desc='Neurons to R',
                                  leave=config.pbar_leave,
                                  disable=config.pbar_hide)}

        nlist = robjects.ListVector(nlist)
        nlist.rownames = [str(n.id) for n in x]

        nlist.rclass = robjects.r('c("neuronlist", "list")')

        return nlist

    elif isinstance(x, core.TreeNeuron):
        # Convert units if applicable
        if isinstance(unit_conversion, (float, int)) and \
           not isinstance(unit_conversion, bool):
            x = x * unit_conversion

        # Prepare list of parents -> root node's parent "None" has to be
        # replaced with -1
        parents = x.nodes.parent_id.values
        # should technically be robjects.r('-1L')
        parents[parents == None] = -1  # DO NOT turn this into "parents is None"!

        swc = robjects.DataFrame({'PointNo': robjects.IntVector(x.nodes.node_id.tolist()),
                                  'Label': robjects.IntVector([0] * x.nodes.shape[0]),
                                  'X': robjects.IntVector(x.nodes.x.tolist()),
                                  'Y': robjects.IntVector(x.nodes.y.tolist()),
                                  'Z': robjects.IntVector(x.nodes.z.tolist()),
                                  'W': robjects.FloatVector([w * 2 for w in x.nodes.radius.tolist()]),
                                  'Parent': robjects.IntVector(parents)
                                  })

        soma = utils.make_non_iterable(x.soma)
        soma_id = int(soma) if not isinstance(soma, type(None)) else robjects.r('NULL')

        meta = {}
        if x.has_connectors:
            try:
                prepost = robjects.IntVector(x.connectors['type'].astype(int).tolist())
            except ValueError:
                prepost = robjects.StrVector(x.connectors['type'].tolist())
            except BaseException:
                raise

            meta['connectors'] = robjects.DataFrame({'node_id': robjects.IntVector(x.connectors.node_id.tolist()),
                                                     'connector_id': robjects.IntVector(x.connectors.connector_id.tolist()),
                                                     'prepost': prepost,
                                                     'x': robjects.IntVector(x.connectors.x.tolist()),
                                                     'y': robjects.IntVector(x.connectors.y.tolist()),
                                                     'z': robjects.IntVector(x.connectors.z.tolist())
                                                     })

        # Generate nat neuron - will reroot to soma (I think)
        return nat.as_neuron(swc, origin=soma_id, **meta)
    elif isinstance(x, core.Dotprops):
        # Convert units if applicable
        if isinstance(unit_conversion, (float, int)) and \
           not isinstance(unit_conversion, bool):
            x = x * unit_conversion

        # Generate matrices for points
        points = robjects.r.matrix(robjects.FloatVector(x.points.T.flatten().tolist()),
                                   nrow=x.points.shape[0], ncol=3,
                                   dimnames=[robjects.r('NULL'), ['X', 'Y', 'Z']])
        vect = robjects.r.matrix(robjects.FloatVector(x.vect.T.flatten().tolist()),
                                 nrow=x.vect.shape[0], ncol=3)
        alpha = robjects.FloatVector(x.alpha.tolist())

        rlist = robjects.ListVector({'points': points,
                                     'alpha': alpha,
                                     'vect': vect})

        rlist.slots['labels'] = robjects.r('NULL')
        rlist.slots['k'] = getattr(x, 'k', 0)

        as_dotprops = robjects.r('as.dotprops')

        return as_dotprops(rlist)
    else:
        raise TypeError(f'Unable to convert data of type "{type(x)}" to R neuron.')


def dotprops2py(dp,
                subset: Optional[Union[List[str], List[int]]] = None
                ) -> 'core.Dotprops':
    """LEGACY FUNCTION: Convert dotprops to pandas DataFrame.

    Parameters
    ----------
    dp :        dotprops neuronlist | neuronlistfh
                Dotprops object to convert.
    subset :    list of str | list of indices, optional
                Neuron names or indices.

    Returns
    -------
    core.Dotprops
        Sub
         of pandas DataFrame. Contains dotprops.
        Can be passed to `plotting.plot3d(dotprops)`

    """
    # Check if list is on demand
    if 'neuronlistfh' in cl(dp) and not subset:
        dp = dp.rx(robjects.IntVector([i + 1 for i in range(len(dp))]))
    elif subset:
        indices = [i for i in subset if isinstance(
            i, int)] + [dp.names.index(n) + 1 for n in subset if isinstance(n, str)]
        dp = dp.rx(robjects.IntVector(indices))

    # This only works if inputs is a neuronlist of dotprops
    if 'neuronlist' in cl(dp):
        df = data2py(dp.slots['df'])
        df.reset_index(inplace=True, drop=True)
    else:
        # If single DataFrame, we don't collect any meta information
        df = pd.DataFrame([])
        dp = [dp]

    points = []
    for i in range(len(dp)):
        this_points = pd.concat([data2py(dp[i][0]), data2py(dp[i][2])], axis=1)
        this_points['alpha'] = dp[i][1]
        this_points.columns = ['x', 'y', 'z', 'x_vec', 'y_vec', 'z_vec', 'alpha']
        points.append(this_points)

    df['points'] = points

    return core.Dotprops(df)


def nblast_allbyall(x: 'core.NeuronList',  # type: ignore  # doesn't like n_cores defaults
                    normalized: bool = True,
                    k: int = 5,
                    resample: Optional[int] = None,
                    n_cores: int = os.cpu_count() - 2,
                    use_alpha: bool = False) -> pd.DataFrame:
    """All-by-all NBLAST using R's `nat.nblast::nblast_allbyall`.

    NBLAST is optimized for data in microns. Original nat function can be found
    [here](https://github.com/jefferislab/nat.nblast/).

    Parameters
    ----------
    x :                 NeuronList | nat.neurons
                        (Tree)Neurons to blast. While not strictly necessary,
                        data should be in microns.
    k :                 int, optional
                        Number of nearest neighbors to use for dotprops generation.
                        Only relevant if input data is not already `Dotprops`.
    resample :          int | bool, optional
                        Resampling during dotprops generation. A good value
                        is `1` which means if data is in microns (which it
                        should!) it will be resampled to 1 tangent vector per
                        micron. Only relevant if input data is not already
                        `Dotprops`.
    normalized :        bool, optional
                        If True, matrix is normalized using z-score.
    n_cores :           int, optional
                        Number of cores to use for nblasting. Default is
                        `os.cpu_count() - 2`.
    use_alpha :         bool, optional
                        Emphasizes neurons' straight parts (backbone) over
                        parts that have lots of branches.

    Returns
    -------
    pandas.DataFrame
                        DataFrame containing the results.

    Examples
    --------
    >>> import navis
    >>> from navis.interfaces import r
    >>> nl = navis.example_neurons()
    >>> # Blast against each other (note the division to get to microns)
    >>> scores = r.nblast_allbyall(nl / 1000)

    """
    if n_cores > 1:
        doMC = importr('doMC')
        doMC.registerDoMC(int(n_cores))

    if 'rpy2' in str(type(x)):
        rn = x
    elif isinstance(x, core.NeuronList):
        if x.shape[0] < 2:
            raise ValueError('You have to provide more than a single neuron.')
        # Check if query or targets are not in microns
        # Note this test can return `None` if it can't be determined
        if not _check_microns(x):
            logger.warning('NBLAST is optimized for data in microns and it looks'
                           ' like these neurons are not in microns.')
        rn = neuron2r(x)
    elif isinstance(x, core.BaseNeuron):
        raise ValueError('You have to provide more than a single neuron.')
    else:
        raise ValueError(f'Must provide NeuronList, not "{type(x)}"')

    # Generate dotprops
    logger.info('Generating dotprops for neurons.')
    xdp = _make_R_dotprops(rn,
                           resample=resample,
                           k=k,
                           parallel=n_cores > 1)

    # Calculate scores
    logger.info('Running all-by-all nblast.')
    pbar = 'text' if n_cores <= 1 else "none"
    scores = r_nblast.nblast(xdp, xdp, **{'normalised': normalized,
                                          '.parallel': n_cores > 1,
                                          '.progress': pbar,
                                          'UseAlpha': use_alpha})

    # Generate DataFrame from scores
    res = data2py(scores).T

    return res


def nblast(query: Union['core.TreeNeuron', 'core.NeuronList', 'core.Dotprops'],  # type: ignore  # doesn't like n_cores default
           target: Optional[str] = None,
           scores: Union[Literal['forward'],
                         Literal['mean'],
                         Literal['min'],
                         Literal['max']] = 'forward',
           n_cores: int = os.cpu_count() - 2,
           normalized: bool = True,
           use_alpha: bool = False,
           k: int = 5,
           resample: Optional[int] = None) -> pd.DataFrame:
    """NBLAST using R's `nat.nblast::nblast`.

    Please note that NBLAST is optimized for units in microns.

    Parameters
    ----------
    query :         TreeNeuron | NeuronList | Dotprops | nat.neuron
                    Query neuron(s) to NBLAST against the targets.
    target :        TreeNeuron | NeuronList | Dotprops | nat.neuron | str
                    Neuron(s) to run the query against. If a str, must be
                    either:

                        1. the name of a file in `'flycircuit.datadir'`
                        2. a path (e.g. `'.../gmrdps.rds'`)
                        3. an R file object (e.g. `robjects.r("load('.../gmrdps.rds')")`)
                        4. a URL to load the list from (e.g. `'http://.../gmrdps.rds'`)
                        5. "flycircuit"

    scores :        'forward' | 'mean' | 'min' | 'max'
                    Determines the final scores:

                        - 'forward' (default) returns query->target scores
                        - 'mean' returns the mean of query->target and target->query scores
                        - 'min' returns the minium between query->target and target->query scores
                        - 'max' returns the maximum between query->target and target->query scores

    n_cores :       int, optional
                    Number of cores to use for nblasting. Default is
                    `os.cpu_count() - 2`.
    use_alpha :     bool, optional
                    Emphasizes neurons' straight parts (backbone) over parts
                    that have lots of branches.
    normalized :    bool, optional
                    Whether to return normalized NBLAST scores.
    k :             int, optional
                    Number of nearest neighbors to use for dotprops generation.
                    Only relevant if input data is not already `Dotprops`.
    resample :      int | bool, optional
                    Resampling during dotprops generation. A good value
                    is `1` which means if data is in microns (which it
                    should!) it will be resampled to 1 tangent vector per
                    micron. Only relevant if input data is not already
                    `Dotprops`.

    Returns
    -------
    scores :        pandas.DataFrame
                    Matrix with NBLAST scores. Rows are query neurons, columns
                    are targets.

    Examples
    --------
    NBLAST example neurons against flycircuit DB

    >>> import navis
    >>> from navis.interfaces import r
    >>> nl = navis.example_neurons(n=5)
    >>> nl.units
    array([8, 8, 8, 8, 8]) <Unit('nanometer')>
    >>> # Convert to microns
    >>> nl_um = nl / 1000
    >>> # Run the nblast
    >>> scores = r.nblast(nl_um)

    """
    utils.eval_param(scores,
                     name='scores',
                     allowed_values=('forward', 'mean', 'min', 'max'))

    if n_cores > 1:
        doMC = importr('doMC')
        doMC.registerDoMC(int(n_cores))
    #_ = robjects.r(f'registerDoMC({int(n_cores)})')

    # Check if query or targets are not in microns
    # Note this test can return `None` if it can't be determined
    if _check_microns(query) is False:
        logger.warning('NBLAST is optimized for data in microns and it looks '
                       'like your queries are not in microns.')
    if _check_microns(target) is False:
        logger.warning('NBLAST is optimized for data in microns and it looks '
                       'like your targets are not in microns.')

    # Turn query into dotprops
    try:
        query_dps = _make_R_dotprops(query,
                                     resample=resample,
                                     k=k,
                                     parallel=n_cores > 1)
    except NotImplementedError:
        raise NotImplementedError('Unable to produce R dotprops for query of '
                                  f'type {type(query)}')

    # Turn target into dotprops
    try:
        target_dps = _make_R_dotprops(target,
                                      resample=resample,
                                      k=k,
                                      parallel=n_cores > 1)
    except NotImplementedError:
        raise NotImplementedError('Unable to produce R dotprops for target of '
                                  f'type {type(target)}')

    # Make sure dotprops are in neuronlists
    if 'neuronlist' not in cl(query_dps):
        logger.info('Generating dotprops for query neurons.')
        target_dps = nat.neuronlist(query_dps)

    if 'neuronlist' not in cl(target_dps):
        logger.info('Generating dotprops for target neurons.')
        target_dps = nat.neuronlist(target_dps)

    logger.info('Running nblast.')
    pbar = 'text' if n_cores <= 1 else "none"
    sc = r_nblast.nblast(query_dps, target_dps,
                         **{'normalised': normalized,
                            '.parallel': n_cores > 1,
                            '.progress': pbar,
                            'UseAlpha': use_alpha})

    # Generate DataFrame from scores
    res = data2py(sc)

    # When nblasting with only 1 query or 1 target neuron the data
    # returned will be a dictionary {'ID1.ID2': score, ...}
    if isinstance(res, dict):
        # Turn dict into DataFrame
        res = _parse_nblast_dict(res)

    # Transpose so that query neurons are rows
    res = res.T

    # Calculate reverse scores
    if scores != 'forward':
        logger.info('Running reverse nblast.')
        scr = r_nblast.nblast(target_dps, query_dps,
                              **{'normalised': normalized,
                                 '.parallel': n_cores > 1,
                                 '.progress': pbar,
                                 'UseAlpha': use_alpha})

        scr = data2py(scr)
        if isinstance(scr, dict):
            scr = _parse_nblast_dict(scr)

        # Make 100% sure that the order is correct
        scr = scr.loc[res.index, res.columns]

        if scores == 'mean':
            res = (res + scr) / 2
        elif scores == 'min':
            res.loc[:, :] = np.dstack((res.values, scr.values)).min(axis=2)
        elif scores == 'max':
            res.loc[:, :] = np.dstack((res.values, scr.values)).max(axis=2)

    return res


def _parse_nblast_dict(x):
    """Parse dict of NBLAST results."""
    # Parsing dict of this format: {'ID1.ID2': score, ...}
    edges = pd.DataFrame([[*k.split('.'), v] for k, v in x.items()],
                         columns=['query', 'target', 'score'])
    mat = edges.pivot(index='target', columns='query', values='score')
    mat.index.name, mat.columns.name = None, None
    return mat


def _check_microns(x, warn=True):
    """Check if neuron data is in microns.

    Returns either [True, None (=unclear), False]
    """
    if isinstance(x, core.NeuronList):
        check = np.array([_check_microns(n) for n in x])
        if np.all(check):
            return True
        # Do NOT change the "check == False" to "check is False" here!
        elif np.any(check == False):
            return False
        return None

    u = getattr(x, 'units', None)
    if isinstance(u, (config.ureg.Quantity, config.ureg.Unit)):
        if not u.unitless:
            return u.to_compact().units == config.ureg.Unit('um')

    return None


def _make_R_dotprops(x, k=5, resample=None, parallel=False):
    """Try to make dotprops from input data."""
    if isinstance(resample, type(None)) or not resample:
        resample = robjects.r('NA')

    # If x is string it might be a flycircuit data type
    if isinstance(x, str) or x is None:
        try:
            flycircuit = importr('flycircuit')
            datadir = robjects.r('getOption("flycircuit.datadir")')[0]
        except BaseException:
            logger.error('R Flycircuit not found.')

    # Fallback are the flycircuit dotprops
    if x is None:
        if not os.path.isfile(datadir + '/dpscanon.rds'):
            raise ValueError('Unable to find default DPS database dpscanon.rds '
                             'in flycircuit.datadir. Please provide database '
                             'using target parameter.')
        logger.info('DPS database not explicitly provided. Loading local '
                    'FlyCircuit DB from dpscanon.rds')
        x = robjects.r(f'read.neuronlistfh("{datadir}/dpscanon.rds")')

    # If string, try to load the R object
    if isinstance(x, str):
        if x.startswith('http') or '/' in x:
            x = robjects.r(f'read.neuronlistfh("{x}")')
        else:
            x = robjects.r(f'read.neuronlistfh("{datadir}/{x}")')

    # If navis neuron convert to R neuron/list
    if isinstance(x, (core.BaseNeuron, core.NeuronList)):
        x = neuron2r(x)

    # At this point we are expecting an R object
    if 'rpy2' not in str(type(x)):
        raise NotImplementedError(f'Unable to convert target of type {type(x)} into '
                                  'R dotprops.')

    # Convert to dotprops if required
    # Note: this logic should be improved at some point
    if all(['dotprops' in cl(n) for n in x]):
        return x
    else:
        dotprops = robjects.r('dotprops')
        if not parallel:
            dps = dotprops(x, resample=resample, k=k)
        else:
            nlapply = robjects.r('nlapply')
            dps = nlapply(x, dotprops,
                          resample=resample, k=k,
                          **{'.parallel': True, '.progress': 'none'})

    return dps


class NBLASTresults:
    """Class holding NBLAST results and wrappers that allow easy plotting.

    Attributes
    ----------
    results :   pandas.Dataframe
                Contains top N results.
    sc :        Robject
                Contains original RNblast forward scores.
    scr :       Robject
                Original R Nblast reverse scores (Top N only).
    query :     R object
                The original query neurons.
    param :     dict
                Contains parameters used for nblasting.
    target :    R robject
                The original target neurons.
    date :      datetime object
                Time of nblasting.

    Examples
    --------
    >>> import navis
    >>> # Blast neuron by skeleton ID
    >>> nbl = navis.nblast( skid, remote_instance = rm )
    >>> # Sort results by mu_score
    >>> nbl.sort( 'mu_score' )
    >>> # Show table
    >>> nbl.results
    >>> # 3D plot top 5 hits using vispy
    >>> canvas, view = nbl.plot(hits=5)
    >>> # Show distribution of results
    >>> import matplotlib.pyplot as plt
    >>> nbl.results.hist( layout=(3,1), sharex=True)
    >>> plt.show()

    """

    def __init__(self, results, sc, scr, query, target, nblast_param):
        """ Init function."""
        self.results = results  # this is pandas Dataframe holding top N results
        self.sc = sc  # original Nblast forward scores
        self.scr = scr  # original Nblast reverse scores (Top N only)
        self.query = query  # the original query neuron(s)
        self.target = target  # the original target neurons
        self.param = nblast_param  # parameters used for nblasting
        self.date = datetime.now()  # time of nblasting

    def sort(self, columns: Union[str, List[str]]):
        """ Sort results by given column(s)."""
        self.results.sort_values(columns, inplace=True, ascending=False)
        self.results.reset_index(inplace=True, drop=True)

    def plot3d(self,
               hits: int = 5,
               plot_neuron: bool = True,
               plot_brain: bool = True,
               **kwargs):
        """Plot nblast hits using `navis.plot3d()`.

        Parameters
        ----------
        hits :  int | str | list thereof, optional
                Nblast hits to plot (default = 5). Can be:

                1. int: e.g. `hits=5` for top 5 hits
                2 .list of ints: e.g. `hits=[2,5]` to plot hits 2 and 5
                3. string: e.g. `hits='THMARCM-198F_seg1'` to plot this neuron
                4. list of strings: e.g. `['THMARCM-198F_seg1', npfMARCM-'F000003_seg002']` to plot multiple neurons by their gene name

        plot_neuron :   bool
                        If `True`, the nblast query neuron will be plotted.
        plot_brain :    bool
                        If `True`, the reference brain will be plotted.
        **kwargs
                        Parameters passed to [`navis.plot3d`][].
                        See `help(navis.plot3d)` for details.

        Returns
        -------
        Depending on the backends used by `navis.plot3d()`:

        vispy (default) : canvas, view
        plotly : plotly figure dictionary

        You can specify the backend by using e.g. `backend='plotly'` in
        **kwargs. See `help(navis.plot3d)` for details.

        """
        nl = self.get_dps(hits)

        n_py = neuron2py(self.neuron)

        # Create colormap with the query neuron being black
        cmap = {n_py.id: (0, 0, 0)}

        colors = np.linspace(0, 1, len(nl) + 1)
        colors = np.array([hsv_to_rgb(c, 1, 1) for c in colors])
        colors *= 255
        cmap.update({e: colors[i] for i, e in enumerate(nl.names)})

        # Prepare brain
        if plot_brain:
            ref_brain = robjects.r(self.param['reference'][8][0] + '.surf')

            verts = data2py(ref_brain[0])[['X', 'Y', 'Z']].values.tolist()
            faces = data2py(ref_brain[1][0]).values
            faces -= 1  # reduce indices by 1
            faces = faces.tolist()

            volumes = {self.param['reference'][8][0]: {'verts': verts,
                                                       'faces': faces}}
        else:
            volumes = []

        kwargs.update({'color': cmap})

        if nl:
            if plot_neuron is True:
                return plotting.plot3d([n_py, neuron2py(nl), volumes], **kwargs)
            else:
                return plotting.plot3d([neuron2py(nl), volumes], **kwargs)

    def get_dps(self, entries: Union[int, str, List[str], List[int]]):
        """Retrieve dotproducts from DPS database (neuronlistfh) as neuronslist.

        Parameters
        ----------
        entries :   int | str | list thereof, optional
                    Neurons to extract from DPS database. Can be:

                    1. int: e.g. `hits=5` for top 5 hits
                    2 .list of ints: e.g. `hits=[2,5]` to plot hits 2 and 5
                    3. string: e.g. `hits = 'THMARCM-198F_seg1'` to plot this neuron
                    4. list of strings:
                       e.g. `['THMARCM-198F_seg1', npfMARCM-'F000003_seg002']`
                       to plot multiple neurons by their gene name

        Returns
        -------
        neuronlist of dotproduct neurons

        """
        if isinstance(entries, int):
            return self.db.rx(robjects.StrVector(self.results.ix[:entries - 1].gene_name.tolist()))
        elif isinstance(entries, str):
            return self.db.rx(entries)
        elif isinstance(entries, (list, np.ndarray)):
            if isinstance(entries[0], int):
                return self.db.rx(robjects.StrVector(self.results.ix[entries].gene_name.tolist()))
            elif isinstance(entries[0], str):
                return self.db.rx(robjects.StrVector(entries))
        else:
            logger.error('Unable to interpret entries provided. See '
                         'help(NBLASTresults.plot3d) for details.')
            return None


def xform_brain(x: Union['core.NeuronObject', 'pd.DataFrame', 'np.ndarray'],
                source: str,
                target: str,
                fallback: Optional[Literal['AFFINE']] = None,
                bulk: bool = False,
                verbose: bool = False,
                **kwargs) -> Union['core.NeuronObject',
                                   'pd.DataFrame',
                                   'np.ndarray']:
    """Transform 3D data between template brains.

    This is a simple wrapper for `nat.templatebrains:xform_brain`.

    Notes
    -----
    For Neurons only: whether there is a change in units during transformation
    (e.g. nm -> um) is inferred by comparing distances between x/y/z coordinates
    before and after transform. This guesstimate is then used to convert
    `.units` and node radii (for TreeNeurons).

    Parameters
    ----------
    x :         Neuron/List | numpy.ndarray | pandas.DataFrame
                Data to transform. Dataframe must contain `['x', 'y', 'z']`
                columns. Numpy array must be shape `(N, 3)`.
    source :    str
                Source template brain that the data currently is in.
    target :    str
                Target template brain that the data should be transformed into.
    fallback :  None | "AFFINE",
                If "AFFINE", will fall back to affine transformation if CMTK
                transformation fails. Else coordinates of points for which the
                transformation failed (e.g. b/c they are out of bounds), will
                be returned as `None`.
    bulk :      bool | int
                If True or number and input is NeuronList, will xform all
                coordinates in chunks (default=100k) instead of neuron-by-neuron.
                This can be ~2x faster (due to reduced overhead) is very memory
                intensive! If `bulk` is a number will process chunks of
                given size.
    **kwargs
                Keyword arguments passed to `nat.templatebrains:xform_brain`

    Returns
    -------
    same type as `x`
                Copy of input with transformed coordinates.

    """
    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            if bulk:
                xf = x.copy()

                # Collect all spatial data
                xyz = []
                for n in xf:
                    if isinstance(n, core.TreeNeuron):
                        xyz.append(n.nodes[['x', 'y', 'z']].values)
                    elif isinstance(n, core.Dotprops):
                        xyz.append(n.points)
                    elif isinstance(n, core.MeshNeuron):
                        xyz.append(n.vertices)
                    else:
                        raise TypeError("Don't know how to transform neuron of "
                                        f"type '{type(n)}'")

                    if n.has_connectors:
                        xyz.append(n.connectors[['x', 'y', 'z']].values)

                # Combine into big matrix
                xyz = np.vstack(xyz)

                if isinstance(bulk, bool):
                    bulk = 100e3

                # Split into chunks of max 100k coordinates
                n_chunks = math.ceil(xyz.shape[0] / bulk)
                chunks = np.array_split(xyz, n_chunks, axis=0)

                # Xform
                xyz_xf = []
                with config.tqdm(desc='Xforming coordinates',
                                 disable=config.pbar_hide,
                                 leave=config.pbar_leave,
                                 total=xyz.shape[0],
                                 ) as pbar:
                    for c in chunks:
                        c_xf = xform_brain(c,
                                           source=source,
                                           target=target,
                                           fallback=fallback,
                                           verbose=verbose,
                                           **kwargs)
                        xyz_xf.append(c_xf)
                        pbar.update(c_xf.shape[0])
                xyz_xf = np.vstack(xyz_xf)

                # Guess change in spatial units
                sample = min(10e3 / xyz.shape[0], .1)
                change, magnitude = _guess_change(xyz, xyz_xf, sample=sample)

                # Some clean-up
                del xyz, chunks

                # Map xformed coordinates back
                offset = 0
                for n in xf:
                    if isinstance(n, core.TreeNeuron):
                        n.nodes.loc[:, ['x', 'y', 'z']] = xyz_xf[offset:offset + n.n_nodes]
                        offset += n.n_nodes
                        # Fix radius based on our best estimate
                        if 'radius' in n.nodes.columns:
                            n.nodes['radius'] *= 10**magnitude
                    elif isinstance(n, core.Dotprops):
                        n.points = xyz_xf[offset:offset + n.points.shape[0]]
                        # Set tangent vectors and alpha to None so they will be regenerated
                        x._vect = x._alpha = None
                        offset += n.points.shape[0]
                    elif isinstance(n, core.MeshNeuron):
                        n.vertices = xyz_xf[offset:offset + n.vertices.shape[0]]
                        offset += n.vertices.shape[0]

                    if n.has_connectors:
                        n.connectors.loc[:, ['x', 'y', 'z']] = xyz_xf[offset:offset + n.n_connectors]
                        offset += n.n_connectors

                    # Make an educated guess as to whether the units have changed
                    if hasattr(n, 'units') and magnitude != 0:
                        if isinstance(n.units, (config.ureg.Unit, config.ureg.Quantity)):
                            n.units = (n.units / 10**magnitude).to_compact()

                return xf

            xf = []
            for n in config.tqdm(x, desc='Xforming',
                                 disable=config.pbar_hide,
                                 leave=config.pbar_leave):
                xf.append(xform_brain(n,
                                      source=source,
                                      target=target,
                                      fallback=fallback,
                                      verbose=verbose,
                                      **kwargs))
            return x.__class__(xf)

    if not isinstance(x, (core.BaseNeuron, np.ndarray, pd.DataFrame, core.Volume)):
        raise TypeError(f'Unable to transform data of type "{type(x)}"')

    if isinstance(x, core.BaseNeuron):
        xf = x.copy()
        # We will collate spatial data to reduce overhead from calling
        # R's xform_brain
        if isinstance(xf, core.TreeNeuron):
            xyz = xf.nodes[['x', 'y', 'z']].values
        elif isinstance(xf, core.MeshNeuron):
            xyz = xf.vertices
        elif isinstance(xf, core.Dotprops):
            xyz = xf.points
        else:
            raise TypeError(f"Don't know how to transform neuron of type '{type(xf)}'")

        # Add connectors
        if xf.has_connnectors:
            xyz = np.vstack([xyz, xf.connectors[['x', 'y', 'z']].values])

        # Do the xform of all spatial data
        xyz_xf = xform_brain(xyz,
                             source=source,
                             target=target,
                             fallback=fallback,
                             verbose=verbose,
                             **kwargs)

        # Guess change in spatial units
        change, magnitude = _guess_change(xyz, xyz_xf)

        # Map xformed coordinates back
        if isinstance(xf, core.TreeNeuron):
            xf.nodes.loc[:, ['x', 'y', 'z']] = xyz_xf[:xf.n_nodes]
            # Fix radius based on our best estimate
            if 'radius' in xf.nodes.columns:
                xf.nodes['radius'] *= 10**magnitude
        elif isinstance(xf, core.Dotprops):
            xf.points = xyz_xf[:xf.points.shape[0]]
            # Set tangent vectors and alpha to None so they will be regenerated
            xf._vect = xf._alpha = None
        elif isinstance(xf, core.MeshNeuron):
            xf.vertices = xyz_xf[:xf.vertices.shape[0]]

        if xf.has_connectors:
            xf.connectors.loc[:, ['x', 'y', 'z']] = xyz_xf[-xf.connectors.shape[0]:]

        # Make an educated guess as to whether the units have changed
        if hasattr(xf, 'units') and magnitude != 0:
            if isinstance(xf.units, (config.ureg.Unit, config.ureg.Quantity)):
                xf.units = (xf.units / 10**magnitude).to_compact()

        # Fix soma radius if applicable
        if hasattr(xf, 'soma_radius') and isinstance(xf.soma_radius, numbers.Number):
            xf.soma_radius *= 10**magnitude

        return xf
    elif isinstance(x, pd.DataFrame):
        if any([c not in x.columns for c in ['x', 'y', 'z']]):
            raise ValueError('DataFrame must have x, y and z columns.')
        x = x.copy()
        x.loc[:, ['x', 'y', 'z']] = xform_brain(x[['x', 'y', 'z']].values.astype(float),
                                                source=source,
                                                target=target,
                                                fallback=fallback,
                                                verbose=verbose,
                                                **kwargs)
        return x
    elif isinstance(x, core.Volume):
        x = x.copy()
        x.vertices = xform_brain(x.vertices,
                                 source=source,
                                 target=target,
                                 fallback=fallback,
                                 verbose=verbose,
                                 **kwargs)
        return x
    elif x.shape[1] != 3:
        raise ValueError('Array must be of shape (N, 3).')

    if not isinstance(source, str):
        TypeError(f'Expected source of type str, got "{type(source)}"')

    if not isinstance(target, str):
        TypeError(f'Expected target of type str, got "{type(target)}"')

    # We need to convert numpy arrays explicitly
    if isinstance(x, np.ndarray):
        if rpy2.__version_vector__[0] < 3:
            x = numpy2ri.py2ro(x)
        else:
            x = numpy2ri.py2rpy(x)

    xf = nat_templatebrains.xform_brain(x,
                                        sample=source,
                                        reference=target,
                                        FallBackToAffine=fallback == 'AFFINE',
                                        Verbose=verbose,
                                        **kwargs)

    return np.array(xf)


def _guess_change(xyz_before, xyz_after, sample=.1):
    """Guess change in units during xforming."""
    if isinstance(xyz_before, pd.DataFrame):
        xyz_before = xyz_before[['x', 'y', 'z']].values
    if isinstance(xyz_after, pd.DataFrame):
        xyz_after = xyz_after[['x', 'y', 'z']].values

    # Select the same random sample of points in both spaces
    if sample <= 1:
        sample = int(xyz_before.shape[0] * sample)
    rnd_ix = np.random.choice(xyz_before.shape[0], sample, replace=False)
    sample_bef = xyz_before[rnd_ix, :]
    sample_aft = xyz_after[rnd_ix, :]

    # Get pairwise distance between those points
    dist_pre = pdist(sample_bef)
    dist_post = pdist(sample_aft)

    # Calculate how the distance between nodes changed and get the average
    # Note we are ignoring nans - happens e.g. when points did not transform.
    with np.errstate(divide='ignore', invalid='ignore'):
        change = dist_post / dist_pre
    # Drop infinite values in rare cases where nodes end up on top of another
    mean_change = np.nanmean(change[change < np.inf])

    # Find the order of magnitude
    magnitude = round(math.log10(mean_change))

    return mean_change, magnitude


def mirror_brain(x: Union['core.NeuronObject', 'pd.DataFrame', 'np.ndarray'],
                 template: str,
                 mirror_axis: Union[Literal['X'],
                                    Literal['Y'],
                                    Literal['Z']] = 'X',
                 transform: Union[Literal['warp'],
                                  Literal['affine'],
                                  Literal['flip']] = 'warp',
                 via: Optional[str] = None,
                 **kwargs) -> Union['core.NeuronObject',
                                    'pd.DataFrame',
                                    'np.ndarray']:
    """Mirror 3D object along given axixs.

    This is a simple wrapper for `nat.templatebrains:mirror_brain`.

    Parameters
    ----------
    x :             Neuron/List | numpy.ndarray | pandas.DataFrame
                    Data to transform. Dataframe must contain `['x', 'y', 'z']`
                    columns. Numpy array must be shape `(N, 3)`.
    template :      str
                    Source template brain space that the data is in.
    mirror_axis :   'X' | 'Y' | 'Z', optional
                    Axis to mirror.
    transform :     'warp' | 'affine' | 'flip', optional
                    Which kind of transform to use.
    via :           str | None
                    If a template brain (e.g. "FCWB") is given will first
                    transform coordinates into that space, then mirror and
                    transform back. Use this if there is no mirror registration
                    for the original template.
    **kwargs
                    Keyword arguments passed to
                    `nat.templatebrains:mirror_brain`.

    Returns
    -------
    same type as `x`
                Copy of input with transformed coordinates.

    """
    assert transform in ['warp', 'affine', 'flip']
    assert mirror_axis in ['X', 'Y', 'Z']

    # If we go via another brain space
    if via:
        # Xform to "via" space
        xf = xform_brain(x, source=template, target=via)
        # Mirror
        xfm = mirror_brain(xf,
                           template=via,
                           mirror_axis=mirror_axis,
                           transform=transform,
                           via=None,
                           **kwargs)
        # Xform back to original template space
        xfm_inv = xform_brain(xfm, source=via, target=template)
        return xfm_inv

    if isinstance(x, core.NeuronList):
        if len(x) == 1:
            x = x[0]
        else:
            xf = []
            for n in config.tqdm(x, desc='Mirroring',
                                 disable=config.pbar_hide,
                                 leave=config.pbar_leave):
                xf.append(mirror_brain(n,
                                       template=template,
                                       mirror_axis=mirror_axis,
                                       transform=transform,
                                       **kwargs))
            return core.NeuronList(xf)

    if not isinstance(x, (core.BaseNeuron, np.ndarray, pd.DataFrame)):
        raise TypeError(f'Unable to transform data of type "{type(x)}"')

    if isinstance(x, core.BaseNeuron):
        x = x.copy()
        if isinstance(x, core.TreeNeuron):
            x.nodes = mirror_brain(x.nodes,
                                   template=template,
                                   mirror_axis=mirror_axis,
                                   transform=transform,
                                   **kwargs)
        elif isinstance(x, core.Dotprops):
            x.points = mirror_brain(x.points,
                                    template=template,
                                    mirror_axis=mirror_axis,
                                    transform=transform,
                                    **kwargs)
            # Set tangent vectors and alpha to None so they will be regenerated
            x._vect = x._alpha = None
        elif isinstance(x, core.MeshNeuron):
            x.vertices = mirror_brain(x.vertices,
                                      template=template,
                                      mirror_axis=mirror_axis,
                                      transform=transform,
                                      **kwargs)
        else:
            raise TypeError(f"Don't know how to transform neuron of type '{type(x)}'")

        if x.has_connectors:
            x.connectors = mirror_brain(x.connectors,
                                        template=template,
                                        mirror_axis=mirror_axis,
                                        transform=transform,
                                        **kwargs)
        return x
    elif isinstance(x, pd.DataFrame):
        if any([c not in x.columns for c in ['x', 'y', 'z']]):
            raise ValueError('DataFrame must have x, y and z columns.')
        x = x.copy()
        x.loc[:, ['x', 'y', 'z']] = mirror_brain(x[['x', 'y', 'z']].values.astype(float),
                                                 template=template,
                                                 mirror_axis=mirror_axis,
                                                 transform=transform,
                                                 **kwargs)
        return x
    elif x.shape[1] != 3:
        raise ValueError('Array must be of shape (N, 3).')

    if not isinstance(template, str):
        TypeError(f'Expected template of type str, got "{type(template)}"')

    # We need to convert numpy arrays explicitly
    if isinstance(x, np.ndarray):
        if rpy2.__version_vector__[0] < 3:
            x = numpy2ri.py2ro(x)
        else:
            x = numpy2ri.py2rpy(x)

    # It appears we need to fetch the brain template first -> passing the string
    # causes an error
    brain = robjects.r(template)

    xf = nat_templatebrains.mirror_brain(x,
                                         brain=brain,
                                         mirrorAxis=mirror_axis,
                                         transform=transform,
                                         **kwargs)

    return np.array(xf)


def get_brain_template_mesh(x: str) -> core.Volume:
    """Fetch brain surface model from `nat.flybrains`, `flycircuit` or `elmr`.

    Parameters
    ----------
    x :             str
                    Name of template brain. For example: 'FCWB', 'FAFB14' or
                    'JFRC2'.

    Returns
    -------
    navis.Volume

    """
    if not x.endswith('.surf'):
        x += '.surf'

    # Get the brain volume (this is a named list)
    brain = robjects.r(f'{x}')

    # Vertices are simply called vertices
    vertices = data2py(brain[brain.names.index('Vertices')])

    # For some reason faces are called "Regions" -> we're using the "first"
    # region although some meshes might include more than one
    faces = data2py(brain[brain.names.index('Regions')][0])

    # Offset to account for difference in indexing between R (1, 2, 3, ...)
    # and  Python (0, 1, 2, ....)
    faces -= 1

    return core.Volume(vertices=vertices[['X', 'Y', 'Z']].values,
                       faces=faces[['V1', 'V2', 'V3']].values,
                       name=x)


def get_neuropil(x: str, template: str = 'FCWB') -> core.Volume:
    """Fetch given neuropil from `nat.flybrains`, `flycircuit` or `elmr`.

    Parameters
    ----------
    x :             str
                    Name of neuropil.
    template :      'FCWB' | 'FAFB14' | 'JFRC', optional
                    Name of the template brain.

    Returns
    -------
    navis.Volume

    """
    if not template.endswith('NP.surf'):
        template += 'NP.surf'

    # Get the brain volume (this is a named list)
    template = robjects.r(f'{template}')

    reglist = template[template.names.index('RegionList')]  # type: ignore  # confused by R objects

    if x not in reglist:
        raise ValueError('Neuropil not found in this brain. Available '
                         f'regions: {", ".join(reglist)}')

    # Get list of faces for desired region
    regions = template[template.names.index('Regions')]  # type: ignore  # confused by R objects
    faces = data2py(regions[regions.names.index(x)])  # type: ignore  # confused by R objects

    # Offset to account for difference in indexing between R (1, 2, 3, ...)
    # and  Python (0, 1, 2, ....)
    faces -= 1

    # Get pre-defined color for this neuropil
    color = template[template.names.index('RegionColourList')][regions.names.index(x)]  # type: ignore  # confused by R objects

    # Get vertices
    all_vertices = data2py(template[template.names.index('Vertices')])  # type: ignore  # confused by R objects

    # Remove superfluous vertices
    verts_required = np.unique(faces.values)
    this_verts = all_vertices.iloc[verts_required]

    # Reorder and remap - DO NOT use ".index" here!
    new_map = {int(old)-1: new for old, new in zip(this_verts.index,
                                            np.arange(0, this_verts.shape[0]).astype(int))}
    faces['V1'] = faces.V1.map(new_map)
    faces['V2'] = faces.V2.map(new_map)
    faces['V3'] = faces.V3.map(new_map)

    return core.Volume(vertices=this_verts[['X', 'Y', 'Z']].values,
                       faces=faces.values,
                       name=x,
                       color=color)


def load_rda(fp: str, convert: bool = True):
    """Load and convert R data file (.rda).

    This function should be able to deal with the common data types used in the
    natverse.

    Parameters
    ----------
    fp :        str
                Filepath to rda file.
    convert :   bool | function
                If True, will attempt to convert data from R to Python. If
                `convert` is a function, we expect it to accept the raw R data
                and return the converted.

    Returns
    -------
    data

    """
    if not isinstance(fp, str):
        raise TypeError(f'Expected filepath as string, got "{type(fp)}"')

    if not os.path.isfile(fp):
        raise ValueError(f'File not found: {fp}')

    load = robjects.r('load')
    object_names = load(fp)
    data = robjects.r(object_names[0])

    if convert is True:
        data = data2py(data)
    elif callable(convert):
        data = convert(data)

    return data



from setuptools import setup, find_packages
from pathlib import Path
from runpy import run_path

from extreqs import parse_requirement_files

HERE = Path(__file__).resolve().parent

verstr = run_path(str(HERE / "navis" / "__version__.py"))["__version__"]

install_requires, extras_require = parse_requirement_files(
    HERE / "requirements.txt",
)

dev_only = ["test-notebook", "dev", "docs"]
specialized = ["r", "flybrains", "cloud-volume"]
all_dev_deps = []
all_deps = []
for k, v in extras_require.items():
    if k in specialized:
        continue
    all_dev_deps.extend(v)
    if k not in dev_only:
        all_deps.extend(v)

extras_require["all"] = all_deps
extras_require["all-dev"] = all_dev_deps

vispy_req = extras_require["vispy-default"][0]
default_backend = vispy_req.split("[")[1].split("]")[0]
# listed here: https://vispy.org/installation.html#backend-requirements
for alt in [
    "pyglet",
    "pyqt5", "pyqt6",
    "pyside", "pyside2", "pyside6",
    "glfw", "sdl2", "wx", "tk",
]:
    extras_require[f"vispy-{alt}"] = [vispy_req.replace(default_backend, alt)]

with open("README.md") as f:
    long_description = f.read()

setup(
    name='navis',
    version=verstr,
    packages=find_packages(include=["navis", "navis.*"]),
    license='GNU GPL V3',
    description='Neuron Analysis and Visualization library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://navis-org.github.io/navis/',
    project_urls={
     "Documentation": "https://navis-org.github.io/navis/",
     "Source": "https://github.com/navis-org/navis",
     "Changelog": "https://navis-org.github.io/navis/changelog/",
    },
    author='Philipp Schlegel',
    author_email='pms70@cam.ac.uk',
    keywords='Neuron Analysis Visualization Morphometrics Morphology Anatomy Connectivity Transform Neuroscience NBLAST Skeletons SWC neuPrint',
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        # 'Programming Language :: Python :: 3.12',
    ],
    install_requires=install_requires,
    extras_require=dict(extras_require),
    tests_require=extras_require["dev"],
    # CI runs against >=3.8
    python_requires='>=3.9,<4.0',
    zip_safe=False,

    include_package_data=True

)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import h5py
import os
import pickle
import pint
import warnings

import multiprocessing as mp
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Union, Iterable, Dict, Optional, Any

from .. import config, utils, core


class BaseH5Reader(ABC):
    """Reads neurons from HDF5 files."""

    def __init__(self, filepath: str):
        """Initialize."""
        self.filepath = filepath

    def __enter__(self):
        """Open file on enter."""
        self.f = h5py.File(self.filepath, 'r')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close file on exit."""
        self.f.close()

    def check_compatible(self):
        """Check if this reader is compatible with existing data."""
        fmt = self.f.attrs.get('format_spec')
        if not isinstance(fmt, str):
            raise ValueError('No format specifier found for file '
                             f'{self.filepath}')
        elif fmt != f'hnf_v{self.version}':
            raise ValueError('Unexpected format specifier for file '
                             f'{self.filepath}: "{fmt}"')

    @abstractmethod
    def list_neurons(self):
        """List neurons in file."""
        pass

    @abstractmethod
    def read_neurons(self):
        """Read neurons from file."""
        pass

    def read_dataframe(self, group, subset=None, exclude=None,
                       include_attrs=False, skip_hidden=True):
        """Read dataset within a group into a single pandas DataFrame."""
        # List datasets contained in this group
        datasets = [k for k, v in group.items() if isinstance(v, h5py.Dataset)]

        if isinstance(subset, type(None)):
            subset = datasets
        else:
            subset = utils.make_iterable(subset)

        if isinstance(exclude, type(None)):
            exclude = []
        else:
            exclude = utils.make_iterable(exclude)

        # Read the datasets
        df = pd.DataFrame()
        for col in datasets:
            if col not in subset or col in exclude:
                continue
            # Skip hidden datasets
            if col.startswith('.') and skip_hidden:
                continue

            # This is the dataset
            ds = group[col]

            # Easy-peasy if dataset has 1 dimension
            if ds.ndim == 1:
                df[col] = ds[:]
            # 2-dimensions should not happen but we will simply enumerate cols
            elif ds.ndim == 2:
                for i in range(ds.shape[1]):
                    df[f'{col}_{i}'] = ds[:, i]
            else:
                raise ValueError(f'Dataset {col} has more than two dimensions.')

        if include_attrs:
            for k in group.attrs:
                df.attrs[k] = group.attrs[k]

        return df


class H5ReaderV1(BaseH5Reader):
    """Reads neurons from HDF5 files."""

    version = 1

    def list_neurons(self, from_cache=True):
        """List all neurons in file."""
        if from_cache and hasattr(self, 'neurons'):
            return self.neurons

        # Go over all top level groups
        self.neurons = []
        for id, grp in self.f.items():
            # Skip if not a group
            if not isinstance(grp, h5py.Group):
                continue

            self.neurons.append(id)

        return self.neurons

    def read_neurons(self,
                     subset=None,
                     read='mesh->skeleton->dotprops',
                     strict=False,
                     prefer_raw=False,
                     on_error='stop',
                     progress=True,
                     annotations=False,
                     **kwargs):
        """Read neurons from file."""
        assert isinstance(read, str)

        readers = {'mesh': self.read_meshneuron,
                   'skeleton': self.read_treeneuron,
                   'dotprops': self.read_dotprops}

        # If no subset specified, load all neurons
        if isinstance(subset, type(None)):
            subset = self.list_neurons()
        else:
            subset = [id for id in subset if id in self.f]

        neurons = []
        errors = {}

        # We can not simply use disable=True here because of some odd
        # interactions when this is run in multiple processes
        if progress:
            pbar = config.tqdm(desc='Reading',
                               leave=False,
                               disable=config.pbar_hide,
                               total=len(subset))

        try:
            for id in subset:
                # Go over the requested neuron representations
                for rep in read.split(','):
                    # Strip potential whitespaces
                    rep = rep.strip()

                    # Go over priorities
                    for prio in rep.split('->'):
                        # Strip potential whitespaces
                        prio = prio.strip()
                        # If that neuron type is present
                        if prio in self.f[id]:
                            try:
                                # Read neuron
                                read_func = readers[prio]
                                n = read_func(id,
                                              strict=strict,
                                              prefer_raw=prefer_raw,
                                              **kwargs)

                                # Read annotations
                                if not isinstance(annotations, type(None)):
                                    an = self.read_annotations(id, annotations, **kwargs)
                                    for k, v in an.items():
                                        setattr(n, k, v)

                                # Append neuron and break prio loop
                                neurons.append(n)
                            except BaseException as e:
                                errors[id] = str(e)
                                if on_error in ('stop', 'raise'):
                                    raise e
                                elif on_error == 'warn':
                                    warnings.warn(f'Error parsing {prio} for '
                                                  f'neuron {id}: {e}')

                            break

                if progress:
                    pbar.update()
        except BaseException:
            raise
        finally:
            if progress:
                pbar.close()

        return neurons, errors

    def parse_add_attributes(self, grp, base_grp, neuron,
                             subset=None, exclude=['units_nm']):
        """Parse attributes and associate with neuron."""
        if not isinstance(subset, type(None)):
            subset = utils.make_iterable(subset)

        # First parse base (neuron-level) attributes
        attrs = {}
        for k, v in base_grp.attrs.items():
            if not isinstance(subset, type(None)) and k not in subset:
                continue
            elif k in exclude:
                continue
            attrs[k] = v

        # Now parse attributes specific for this representation
        for k, v in grp.attrs.items():
            if not isinstance(subset, type(None)) and k not in subset:
                continue
            elif k in exclude:
                continue
            attrs[k] = v

        # Now add attributes
        for k, v in attrs.items():
            setattr(neuron, k, v)

    def parse_add_datasets(self, grp, neuron, subset=None, exclude=None,
                           skip_hidden=True):
        """Parse datasets and associate with neuron."""
        if not isinstance(subset, type(None)):
            subset = utils.make_iterable(subset)
        if not isinstance(exclude, type(None)):
            exclude = utils.make_iterable(exclude)

        # Now parse attributes specific for this representation
        for k in grp.keys():
            if not isinstance(subset, type(None)) and k not in subset:
                continue
            elif k.startswith('.') and skip_hidden:
                continue
            elif k in exclude:
                continue
            # Do not remove the [:] as it ensures that we get a numpy array
            setattr(neuron, k, grp[k][:])

    def parse_add_units(self, grp, base_grp, neuron):
        """Parse units and associate with neuron."""
        # Check if we have units
        units = grp.attrs.get('units_nm',
                              base_grp.attrs.get('units_nm',
                                                 None))

        if isinstance(units, str):
            # If string, see if pint can parse it
            neuron.units = units
        elif isinstance(units, np.ndarray):
            # Currently navis .units does support x/y/z units
            # We will use only the first entry for now
            neuron.units = f'{units[0]} nm'
        elif not isinstance(units, type(None)):
            neuron.units = f'{units} nm'

    def read_annotations(self, id, annotations, **kwargs):
        """Read annotations for given neuron from file."""
        # Get the group for this neuron
        neuron_grp = self.f[id]

        an_grp = neuron_grp.get('annotations')
        if not an_grp:
            return {}

        if isinstance(annotations, bool):
            annotations = [k for k, grp in an_grp.items() if isinstance(grp, type(h5py.Group))]
        else:
            annotations = utils.make_iterable(annotations)

        parsed_an = {}
        for an in annotations:
            if an in an_grp:
                parsed_an[an] = self.read_dataframe(an_grp[an],
                                                    include_attrs=True)

        return parsed_an

    def read_treeneuron(self, id, strict=False, prefer_raw=False, **kwargs):
        """Read given TreeNeuron from file."""
        # Get the group for this neuron
        neuron_grp = self.f[id]

        # See if this neuron has a skeleton
        if 'skeleton' not in neuron_grp:
            raise ValueError(f'Neuron {id} has no skeleton')
        sk_grp = neuron_grp['skeleton']

        if '.serialized_navis' in sk_grp and not prefer_raw:
            return pickle.loads(sk_grp['.serialized_navis'][()])

        # Parse node table
        nodes = self.read_dataframe(sk_grp,
                                    subset=['node_id', 'parent_id',
                                            'x', 'y', 'z',
                                            'radius'] if strict else None)

        n = core.TreeNeuron(nodes, id=id)

        # Check if we have units
        self.parse_add_units(sk_grp, neuron_grp, n)

        # Parse attributes
        self.parse_add_attributes(sk_grp, neuron_grp, n,
                                  subset=['soma'] if strict else None)

        return n

    def read_dotprops(self, id, strict=False, prefer_raw=False, **kwargs):
        """Read given Dotprops from file."""
        # Get the group for this neuron
        neuron_grp = self.f[id]

        # See if this neuron has dotprops
        if 'dotprops' not in neuron_grp:
            raise ValueError(f'Neuron {id} has no dotprops')
        dp_grp = neuron_grp['dotprops']

        if '.serialized_navis' in dp_grp and not prefer_raw:
            return pickle.loads(dp_grp['.serialized_navis'][()])

        # Parse dotprop arrays
        points = dp_grp['points']
        vect = dp_grp['vect']
        alpha = dp_grp['alpha']
        k = dp_grp.attrs['k']

        n = core.Dotprops(points=points,
                          k=k,
                          vect=vect,
                          alpha=alpha,
                          id=id,
                          name=neuron_grp.attrs.get('neuron_name'))

        # Check if we have units
        self.parse_add_units(dp_grp, neuron_grp, n)

        # Parse additional attributes
        self.parse_add_attributes(dp_grp, neuron_grp, n,
                                  subset=['soma'] if strict else None,
                                  exclude=['units_nm', 'k'])

        # Parse additional datatsets
        if not strict:
            self.parse_add_datasets(dp_grp, n,
                                    exclude=['points', 'vect', 'alpha'])

        return n

    def read_meshneuron(self, id, strict=False, prefer_raw=False, **kwargs):
        """Read given MeshNeuron from file."""
        # Get the group for this neuron
        neuron_grp = self.f[id]

        # See if this neuron has a mesh
        if 'mesh' not in neuron_grp:
            raise ValueError(f'Neuron {id} has no mesh')
        me_grp = neuron_grp['mesh']

        if '.serialized_navis' in me_grp and not prefer_raw:
            return pickle.loads(me_grp['.serialized_navis'][()])

        # Parse node table -> do not remove the [:] as it ensures that we get
        # a numpy array
        verts = me_grp['vertices'][:]
        faces = me_grp['faces'][:]

        n = core.MeshNeuron({'vertices': verts, 'faces': faces},
                            id=id,
                            name=neuron_grp.attrs.get('neuron_name'))

        if 'skeleton_map' in me_grp:
            n.skeleton_map = me_grp['skeleton_map']

        # Check if we have units
        self.parse_add_units(me_grp, neuron_grp, n)

        # Parse additional attributes
        self.parse_add_attributes(me_grp, neuron_grp, n,
                                  subset=['soma'] if strict else None,
                                  exclude=['units_nm'])

        # Parse additional datatsets
        if not strict:
            self.parse_add_datasets(me_grp, n,
                                    exclude=['vertices',
                                             'faces',
                                             'skeleton_map'])

        return n


class BaseH5Writer(ABC):
    """Writes neurons to HDF5 files."""

    version = None

    def __init__(self, filepath: str,  mode='a', **kwargs):
        """Initialize.

        Parameters
        ----------
        filepath :      str
                        Path to HDF5 file.
        mode :          "a" | "w" |
                        Mode in which to open H5 file::

                            w	Create file, truncate if exists
                            a	Read/write if exists, create otherwise

        **kwargs
                        Passed to `h5py.File`.
        """
        self.filepath = filepath
        self.mode = mode
        self.kwargs = kwargs

    def __enter__(self):
        """Open file on enter."""
        self.f = h5py.File(self.filepath, self.mode, **self.kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close file on exit."""
        self.f.close()

    def check_compatible(self):
        """Check if this writer is compatible with existing data."""
        fmt = self.f.attrs.get('format_spec')
        if fmt:
            if not fmt.startswith('hnf_'):
                raise ValueError('Unexpected format specifier for file '
                                 f'{self.filepath}: "{fmt}"')

            ver = fmt.split('_')[-1]
            if ver != 'v1':
                raise ValueError(f'File {self.filepath} appears to contain '
                                 f'data from an incompatible version: "{ver}"')

    def write_dataframe(self, df, group, subset=None, exclude=None, overwrite=False):
        """Write dataframe to group."""
        assert isinstance(df, pd.DataFrame)

        if isinstance(subset, type(None)):
            subset = df.columns
        else:
            subset = utils.make_iterable(subset)

        if isinstance(exclude, type(None)):
            exclude = []
        else:
            exclude = utils.make_iterable(exclude)

        # Remove datasets if it already exists
        for c in df.columns:
            if c not in subset or c in exclude:
                continue

            if c in group:
                if not overwrite:
                    raise ValueError(f'Dataset {c} already exists in group "{group}"')
                del group[c]

            # Convert categoricals
            if isinstance(df[c].dtype, pd.CategoricalDtype):
                data = np.asarray(df[c])
            else:
                data = df[c].values

            # HDF5 does not like numpy strings ("<U4") or object
            if data.dtype.type in (np.str_, np.object_):
                data = data.astype("S")

            group.create_dataset(c, data=data, compression='gzip')

    def write_base_info(self):
        """Add version and url to root path."""
        self.f.attrs['format_spec'] = f'hnf_v{self.version}'
        self.f.attrs['format_url'] = 'https://github.com/navis-org/navis'

    @abstractmethod
    def write_neurons(self, neuron: 'core.NeuronObject'):
        """Write Neuron to file."""
        pass


class H5WriterV1(BaseH5Writer):
    """Implements v1 of the HDF schema.

    See <url> for format specs.
    """

    version = 1

    def write_neurons(self, neuron, serialized=True, raw=False,
                      overwrite=True, annotations=None, **kwargs):
        """Write neuron to file."""
        if isinstance(neuron, core.NeuronList):
            for n in config.tqdm(neuron, desc='Writing',
                                 leave=False,
                                 disable=config.pbar_hide):
                self.write_neurons(n, overwrite=overwrite,
                                   annotations=annotations, **kwargs)
            return

        if isinstance(neuron, core.TreeNeuron):
            self.write_treeneuron(neuron, serialized=serialized, raw=raw,
                                  overwrite=overwrite, **kwargs)
        elif isinstance(neuron, core.MeshNeuron):
            self.write_meshneuron(neuron, serialized=serialized, raw=raw,
                                  overwrite=overwrite, **kwargs)
        elif isinstance(neuron, core.Dotprops):
            self.write_dotprops(neuron, serialized=serialized, raw=raw,
                                overwrite=overwrite, **kwargs)
        else:
            raise TypeError(f'Unable to write object of type "{type(neuron)}"'
                            'to HDF5 file.')

        # Write annotations
        if not isinstance(annotations, type(None)):
            _ = self.write_annotations(neuron, annotations,
                                       overwrite=overwrite,
                                       **kwargs)

    def write_annotations(self, neuron, annotations, overwrite=True, **kwargs):
        """Write annotations for given neuron to file."""
        # Get this neuron's group
        neuron_grp = self.get_neuron_group(neuron)
        # Create annotation group if it does not exist
        an_grp = neuron_grp.require_group('annotations')

        annotations = utils.make_iterable(annotations)

        for an in annotations:
            data = getattr(neuron, an)
            if isinstance(data, pd.DataFrame):
                data_grp = an_grp.require_group(an)
                self.write_dataframe(data, data_grp, overwrite=overwrite)
            else:
                raise ValueError(f'Unable to write "{an}" of type '
                                 f'"({type(data)})" to HDF5 file.')

    def get_neuron_group(self, neuron):
        """Get group for neuron or create if it does not exists."""
        # Can only use strings as group names
        id = str(neuron.id)

        grp = self.f.require_group(id)

        # Write some basic info
        if hasattr(neuron, 'name'):
            grp.attrs['neuron_name'] = neuron.name

        return grp

    def write_treeneuron(self, neuron,
                         serialized=True, raw=False,
                         overwrite=True, **kwargs):
        """Write TreeNeuron to file."""
        assert isinstance(neuron, core.TreeNeuron)

        # Get the group for this neuron -> this also write BaseInfo
        neuron_grp = self.get_neuron_group(neuron)

        # See if this neuron already has a skeleton
        if 'skeleton' in neuron_grp:
            if not overwrite:
                raise ValueError('File already contains a skeleton for neuron '
                                 f'{neuron.id}')
            del neuron_grp['skeleton']

        sk_grp = neuron_grp.require_group('skeleton')

        if serialized:
            sk_grp.create_dataset('.serialized_navis',
                                  data=np.void(pickle.dumps(neuron)))

        if raw:
            # Write info
            units = neuron_nm_units(neuron)
            if units:
                sk_grp.attrs['units_nm'] = units

            if neuron.has_soma:
                sk_grp.attrs['soma'] = neuron.soma

            # Write node table
            self.write_dataframe(neuron.nodes,
                                 group=sk_grp,
                                 exclude=['type'],
                                 overwrite=overwrite)

    def write_dotprops(self, neuron,
                       serialized=True, raw=False,
                       overwrite=True, **kwargs):
        """Write Dotprops to file."""
        assert isinstance(neuron, core.Dotprops)

        # Get the group for this neuron -> this also write BaseInfo
        neuron_grp = self.get_neuron_group(neuron)

        # See if this neuron already has dotprops
        if 'dotprops' in neuron_grp:
            if not overwrite:
                raise ValueError('File already contains dotprops for neuron '
                                 f'{neuron.id}')
            del neuron_grp['dotprops']

        dp_grp = neuron_grp.require_group('dotprops')

        if serialized:
            dp_grp.create_dataset('.serialized_navis',
                                  data=np.void(pickle.dumps(neuron)))

        if raw:
            # Write info
            dp_grp.attrs['k'] = neuron.k
            units = neuron_nm_units(neuron)
            if units:
                dp_grp.attrs['units_nm'] = units
            if neuron.has_soma:
                dp_grp.attrs['soma'] = neuron.soma

            # Write data
            for d in ['points', 'vect', 'alpha']:
                # The neuron really ought to have these but just in case
                if not hasattr(neuron, d):
                    continue

                data = getattr(neuron, d)
                dp_grp.create_dataset(d, data=data, compression='gzip')

    def write_meshneuron(self, neuron,
                         serialized=True, raw=False,
                         overwrite=True, **kwargs):
        """Write MeshNeuron to file."""
        assert isinstance(neuron, core.MeshNeuron)

        # Get the group for this neuron -> this also write BaseInfo
        neuron_grp = self.get_neuron_group(neuron)

        # See if this neuron already has a mesh
        if 'mesh' in neuron_grp:
            if not overwrite:
                raise ValueError('File already contains a mesh for neuron '
                                 f'{neuron.id}')
            del neuron_grp['mesh']

        me_grp = neuron_grp.require_group('mesh')

        if serialized:
            me_grp.create_dataset('.serialized_navis',
                                  data=np.void(pickle.dumps(neuron)))

        if raw:
            # Write info
            units = neuron_nm_units(neuron)
            if units:
                me_grp.attrs['units_nm'] = units
            if neuron.has_soma:
                me_grp.attrs['soma'] = neuron.soma

            # Write data
            for d in ['vertices', 'faces', 'skeleton_map']:
                if not hasattr(neuron, d):
                    continue

                data = getattr(neuron, d)
                me_grp.create_dataset(d, data=data, compression='gzip')


def read_h5(filepath: str,
            read='mesh->skeleton->dotprops',
            subset=None,
            prefer_raw=False,
            annotations=True,
            strict=False,
            reader='auto',
            on_error='stop',
            ret_errors=False,
            parallel='auto') -> 'core.NeuronObject':
    """Read Neuron/List from Hdf5 file.

    This import is following the schema specified
    [here](https://github.com/flyconnectome/hnf)

    Parameters
    ----------
    filepath :          filepath
                        Path to HDF5 file.
    read :              str
                        The HDF5 file might contain skeleton, dotprops and/or
                        mesh representations for any given neuron. This
                        parameter determines which one are returned. Some
                        illustrative examples:

                          - 'mesh', 'skeleton' or 'dotprops' will return only
                            the given representation
                          - 'mesh->skeleton->dotprops' will return a mesh if the
                            neuron has one, a skeleton if it does not and
                            dotprops if it has neither mesh nor skeleton
                          - 'mesh,skeleton,dotprops' will return all available
                            representations
                          - 'mesh,dotprops' will only return meshes and dotprops
                          - 'mesh,skeleton->dotprops' will return the mesh
                            and a skeleton or alternatively the dotprops

                        Note that neurons which have none of the requested
                        representations are silently skipped!
    subset :            list of IDs | slice
                        If provided, will read only a subset of neurons from the
                        file. IDs that don't exist are silently ignored. Also
                        note that due to HDF5 restrictions numeric IDs will be
                        converted to strings.
    prefer_raw :        bool
                        If True and a neuron is saved as both serialized and
                        raw data, will load the neuron from the raw data.
    parallel :          "auto" | bool | int
                        Defaults to `auto` which means only use parallel
                        processing if more than 200 neurons are imported.
                        Spawning and joining processes causes overhead and is
                        considerably slower for imports of small numbers of
                        neurons. Integer will be interpreted as the
                        number of cores (otherwise defaults to
                        `os.cpu_count() - 2`).
    on_error :          "stop" | "warn" | "ignore"
                        What to do if a neuron can not be parsed: "stop" and
                        raise an exception, "warn" and keep going or silently
                        "ignore" and skip.
    ret_errors :        bool
                        If True, will also return a list of errors encountered
                        while parsing the neurons.

    Only relevant for raw data:

    annotations :       bool | str | list of str
                        Whether to load annotations associated with the
                        neuron(s):

                         - `True` reads all annotations
                         - `False` reads no annotations
                         - e.g. `["connenctors"]` reads only "connectors"

                        Non-existing annotations are silently ignored!
    strict :            bool
                        If True, will read only the attributes/columns which
                        are absolutely required to construct the respective
                        neuron representation. This is useful if you either want
                        to keep memory usage low or if any additional attributes
                        are causing troubles. If False (default), will read
                        every attribute and dataframe column and attach it to
                        the neuron.
    reader :            "auto" | str | subclass of BaseH5Reader
                        Which reader to use to parse the given format. By
                        default ("auto") will try to pick the correct parser
                        for you depending on the `format_spec` attribute in
                        the HDF5 file. Alternatively, you can also provided either
                        a format version (e.g. "v1") or a subclass of BaseH5Reader
                        that is capable of reading neurons from the file.

    Returns
    -------
    neurons :           navis.NeuronList

    errors :            dict
                        If `ret_errors=True` return dictionary with errors:
                        `{id: "error"}`.

    Examples
    --------
    See [`navis.write_h5`][] for examples.


    See Also
    --------
    [`navis.write_h5`][]
                        Write neurons to HDF5 file.
    [`navis.io.inspect_h5`][]
                        Extract meta data (format, number of neurons,
                        available annotations and representations) from
                        HDF5 file. This is useful if you don't know what's
                        actually contained within the HDF5 file.

    """
    utils.eval_param(read, name='read', allowed_types=(str, ))
    utils.eval_param(on_error, name='on_error',
                     allowed_values=('stop', 'warn', 'ignore'))

    # Make sure the read string is "correct"
    for rep in read.split(','):
        rep = rep.strip()
        for prio in rep.split('->'):
            prio = prio.strip()
            if prio not in ('mesh', 'skeleton', 'dotprops'):
                raise ValueError(f'Unexpected representation in `read` parameter: {prio}')

    # Get info for this file
    filepath = os.path.expanduser(filepath)
    info = inspect_h5(filepath, inspect_neurons=True, inspect_annotations=False)

    # Get a reader for these specs
    if reader == 'auto':
        if info['format_spec'] is None:
            config.logger.warning(
                'No format specifier found in file, suggesting this file may not have '
                'been created using NAVis. We will try to read using the latest '
                'version of the schema. If this fails you may have to specify a reader '
                'or version manually (see the `reader` parameter).')
            reader = READERS['latest']
        elif info['format_spec'] not in READERS:
            raise TypeError(f'No reader for HDF5 format {info["format_spec"]}')
        reader = READERS[info['format_spec']]
    elif isinstance(reader, str):
        if reader not in READERS:
            raise TypeError(f'No reader for HDF5 format "{reader}"')
        reader = READERS[reader]
    elif not isinstance(reader, BaseH5Reader):
        raise TypeError('If provided, the reader must be a subclass of '
                        f'BaseH5Reader - got "{type(reader)}"')

    # By default only use parallel if there are more than 200 neurons
    if parallel == 'auto':
        if len(info['neurons']) > 200:
            parallel = True
        else:
            parallel = False

    if not parallel:
        # This opens the file
        with reader(filepath) as r:
            nl, errors = r.read_neurons(subset=subset,
                                        read=read,
                                        strict=strict,
                                        on_error=on_error,
                                        annotations=annotations)
    else:
        # Do not swap this as `isinstance(True, int)` returns `True`
        if isinstance(parallel, (bool, str)):
            n_cores = os.cpu_count() - 2
        else:
            n_cores = int(parallel)

        # If subset not specified, fetch all neurons
        if isinstance(subset, type(None)):
            subset = list(info['neurons'])
        elif isinstance(subset, slice):
            subset = list(info['neurons'])[subset]
        else:
            # Make sure it's an iterable and strings
            subset = utils.make_iterable(subset).astype(str)

        # Just to leave note that I tried splitting the array into
        # `n_cores` chunks but that caused massive memory usage in the
        # spawned processes without being any faster - reading and returning
        # one neuron at a time seems to be the most efficient way
        reader = READERS[info['format_spec']]
        with mp.Pool(processes=n_cores) as pool:
            futures = pool.imap(_h5_reader_worker, [dict(reader=reader,
                                                         filepath=filepath,
                                                         read=read,
                                                         strict=strict,
                                                         prefer_raw=prefer_raw,
                                                         on_error=on_error,
                                                         annotations=annotations,
                                                         subset=[x]) for x in subset],
                                chunksize=1)

            # Wait for results and show progress bar althewhile
            # Do not close the pool before doing this
            res = list(config.tqdm(futures,
                                   desc='Reading',
                                   total=len(subset)))

        # Unpack results
        nl = []
        errors = {}
        for n, e in res:
            nl += n
            errors.update(e)

        # Warnings will not have propagated
        if on_error == 'warn':
            for e in errors:
                warnings.warn(f"Error reading neuron {e}: {errors[e]}")

    if ret_errors:
        return core.NeuronList(nl), errors
    else:
        return core.NeuronList(nl)


def _h5_reader_worker(kwargs):
    """Trigger reading neurons from H5 file."""
    reader = kwargs['reader']
    filepath = kwargs['filepath']
    read = kwargs['read']
    subset = kwargs['subset']
    annotations = kwargs['annotations']
    strict = kwargs['strict']
    on_error = kwargs['on_error']
    prefer_raw = kwargs['prefer_raw']
    # This opens the file
    with reader(filepath) as r:
        return r.read_neurons(subset=subset,
                              read=read,
                              strict=strict,
                              on_error=on_error,
                              prefer_raw=prefer_raw,
                              progress=False,
                              annotations=annotations)


def write_h5(n: 'core.NeuronObject',
             filepath: str,
             serialized: bool = True,
             raw: bool = False,
             annotations: Optional[Union[str, list]] = None,
             format: str = 'latest',
             append: bool = True,
             overwrite_neurons: bool = False) -> 'core.NeuronObject':
    """Write Neuron/List to Hdf5 file.

    Parameters
    ----------
    n :                 Neuron | NeuronList
                        Neuron(s) to write to file.
    filepath :          str
                        Path to HDF5 file. Will be created if it does not
                        exist. If it does exist, we will add data to it.
    serialized :        bool
                        Whether to write a serialized (pickled) version of the
                        neuron to file.
    raw :               bool
                        Whether to write the neurons' raw data to file. This
                        is required to re-generate neurons from tools other
                        than `navis` (e.g. R's `nat`). This follows the schema
                        specified [here](https://github.com/flyconnectome/hnf).
    append :            bool
                        If file already exists, whether to append data or to
                        overwrite the entire file.
    overwrite_neurons : bool
                        If a given neuron already exists in the h5 file whether
                        to overwrite it or throw an exception.

    Only relevant if `raw=True`:

    annotations :       str | list thereof, optional
                        Whether to write annotations (e.g. "connectors")
                        associated with the neuron(s) to file. Annotations
                        must be pandas DataFrames. If a neuron does not contain
                        a given annotation, it is silently skipped.
    format :            "latest" | "v1"
                        Which version of the format specs to use. By default
                        use latest. Note that we don't allow mixing format
                        specs in the same HDF5 file. So if you want to write
                        to a file which already contains data in a given
                        format, you have to use that format.

    Returns
    -------
    Nothing

    Examples
    --------
    >>> import navis
    >>> # First get mesh, skeleton and dotprop representations for some neurons
    >>> sk = navis.example_neurons(5, kind='skeleton')
    >>> me = navis.example_neurons(5, kind='mesh')
    >>> dp = navis.make_dotprops(sk, k=5)
    >>> # Write them to a file
    >>> navis.write_h5(sk + me + dp, '~/test.h5', overwrite_neurons=True)
    >>> # Read back from file
    >>> nl = navis.read_h5('~/test.h5')

    See Also
    --------
    [`navis.read_h5`][]
                        Read neurons from h5 file.

    """
    if not serialized and not raw:
        raise ValueError('`serialized` and `raw` must not both be False.')

    utils.eval_param(format, name='format',
                     allowed_values=tuple(WRITERS.keys()))

    filepath = os.path.expanduser(filepath)

    # Get the writer for the specified format
    writer = WRITERS[format]

    # This opens the file
    with writer(filepath, mode='a' if append else 'w') as w:
        w.write_base_info()
        w.write_neurons(n,
                        raw=raw,
                        serialized=serialized,
                        overwrite=overwrite_neurons,
                        annotations=annotations)


def inspect_h5(filepath, inspect_neurons=True, inspect_annotations=True):
    """Extract basic info from Hdf5 file.

    Parameters
    ----------
    filepath :              str
                            Path to HDF5 file.
    inspect_neurons :       bool
                            If True, will return info about the neurons contained.
    inspect_annotations :   bool
                            If True, include info about annotations associated
                            with each neuron.

    Returns
    -------
    dict
                        Returns a dictionary with basic info about the file.
                        An example::

                         {
                          'format_spec': 'hnf_v1', # format specifier
                          'neurons': {
                                      'someID': {'skeleton': True,
                                                 'mesh': False,
                                                 'dotprops': True,
                                                 'annotations': ['connectors']},
                                      'someID2': {'skeleton': False,
                                                  'mesh': False,
                                                  'dotprops': True,
                                                  'annotations': ['connectors']}
                                     }
                            }


    """
    if not isinstance(filepath, str):
        raise TypeError(f'`filepath` must be str, got "{type(filepath)}"')

    if not os.path.isfile(filepath):
        raise ValueError(f'{filepath} does not exist')

    info = dict()
    with h5py.File(filepath, 'r') as f:
        info['format_spec'] = f.attrs.get('format_spec')
        # R strings are automatically stored as vectors
        info['format_spec'] = utils.make_non_iterable(info['format_spec'])

        if inspect_neurons:
            info['neurons'] = {}
            # Go over all top level groups
            for id, grp in f.items():
                # Skip if not a group
                if not isinstance(grp, h5py.Group):
                    continue

                # Do not change this test
                this = {}
                if 'skeleton' in grp:
                    this['skeleton'] = True
                if 'mesh' in grp:
                    this['mesh'] = True
                if 'dotprops' in grp:
                    this['dotprops'] = True

                if this:
                    info['neurons'][id] = this
                    if inspect_annotations:
                        annotations = grp.get('annotations', None)
                        if annotations:
                            info['neurons'][id]['annotations'] = list(annotations.keys())

    return info


def neuron_nm_units(neuron):
    """Return neuron's units in nanometers.

    Returns `None` if units are dimensionless.

    """
    units = getattr(neuron, 'units')

    if isinstance(units, type(None)):
        return

    if not isinstance(units, (pint.Quantity, pint.Unit)):
        return None

    # Casting to nm throws an error if dimensionless
    try:
        return units.to('nm').magnitude
    except pint.DimensionalityError:
        return None
    except BaseException:
        raise


WRITERS = {'v1': H5WriterV1,
           'latest': H5WriterV1}

READERS = {'hnf_v1': H5ReaderV1,
           'latest': H5ReaderV1}
